# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import datetime as datetime_module
from typing import Iterable, Mapping, Union, Type, Optional, Any
import itertools
import collections.abc

from ...jamming import BaseJamDatabase, Jam, JamId, JamItem, JamKind, JamParchment
from ..core import Swank, SwankDatabase, SwankRef
from .base_field import BaseField
from .. import utils


class SwankField(BaseField):

    field_type_name = 'swank'

    def from_jam(self, jam: Jam, swank_database: SwankDatabase, *,
                 swank_cache: dict) -> Optional[Swank]:
        if jam is None:
            return None
        else:
            jam_kind_name, jam_id_name, jam_index = jam
            return SwankRef(swank_database, utils.name_to_type(jam_kind_name),
                           JamId(jam_id_name), jam_index)

    def to_jam(self, value: Swank, swank_database: SwankDatabase) -> Jam:
        if value is None:
            return None
        else:
            assert value.has_jam_id_and_index
            swank_ref = SwankRef.from_swank_or_ref(value)
            return (swank_ref.jam_kind_name, str(swank_ref.jam_id), swank_ref.jam_index)

    def get_default_value(self, swank_database: SwankDatabase) -> None:
        return None

    def __get__(self, swank: Optional[Swank], swank_type: Optional[Type] = None) -> Optional[Swank]:
        if swank is None:
            assert swank_type is not None
            return self
        swank_ref: Optional[SwankRef] = swank._Swank__field_values[self.name]
        if swank_ref is None:
            return None
        else:
            return swank_ref.get()

    def __set__(self, swank: Swank,
                other_swank_or_ref: Optional[Union[Swank, SwankRef]]) -> None:
        if other_swank_or_ref is None:
            swank._Swank__field_values[self.name] = None
        else:
            swank._Swank__field_values[self.name] = \
                                             SwankRef.from_swank_or_ref(other_swank_or_ref)



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import functools
import warnings
import re
import json
import urllib.parse
import threading
import pathlib
import requests
import logging
from typing import (Optional, Tuple, Union, Container, Hashable, Iterator,
                    Iterable, Any, Dict, FrozenSet, Callable, Type, Sequence)

import flask
import flask_cors

from marley import constants
from marley.worlds.grid_royale.core import get_games_folder
from marley.jamswank.jamming import server as jamming_server
from marley.jamswank.jamming.jam_file_database import JamFileDatabase
from marley.jamswank.swanking import EmptyJam

from .user_sessioning import ArezzoUserSession


DEFAULT_HOST = '127.0.0.1'
DEFAULT_PORT = 22390
DIST_FOLDER = pathlib.Path(__file__).parent / 'arezzo' / 'dist'

assert DIST_FOLDER.exists()
assert DIST_FOLDER.is_dir()

def make_url_from_host_port(host: str, port: Union[str, int]):
    return urllib.parse.urlunsplit(('http', f'{host}:{port}', '', '', ''))

DEFAULT_URL = make_url_from_host_port(DEFAULT_HOST, DEFAULT_PORT)


app = flask.Flask('arezzo', static_url_path='/', static_folder=str(DIST_FOLDER))
flask_cors.CORS(app)
jamming_server.add_to_flask_app(constants.jam_file_database, app)

@app.errorhandler(404)
def page_not_found(exception) -> tuple[str, int]:
    return ((DIST_FOLDER / 'index.html').read_text(), 200)



@app.route('/api/shutdown', methods=('POST',))
def shutdown() -> str:
    shutdown_function = flask.request.environ.get('werkzeug.server.shutdown')
    if shutdown_function is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    shutdown_function()
    return 'Shut down flask server.'

@app.route('/api/favorites', methods=('GET',))
def get_favorites():
    try:
        arezzo_user_session = ArezzoUserSession.load(constants.swank_database)
    except IndexError:
        return json.dumps([])
    else:
        return json.dumps(arezzo_user_session.favorites)

@app.route('/api/favorites/<path:favorite>', methods=('POST',))
def add_favorite(favorite):
    if not re.fullmatch(r'^[a-zA-Z0-9._/\[\]*]+$', favorite):
        return flask.Response(status=400)
    with ArezzoUserSession.lock_and_load_or_create(constants.swank_database,
                                                   save=True) as arezzo_user_session:
        arezzo_user_session.favorites = sorted(set(arezzo_user_session.favorites) |
                                               {favorite})
    return ''

@app.route('/api/favorites/<path:favorite>', methods=('DELETE',))
def delete_favorite(favorite):
    if not re.fullmatch(r'^[a-zA-Z0-9._/\[\]*]+$', favorite):
        return flask.Response(status=400)
    with ArezzoUserSession.lock_and_load_or_create(constants.swank_database,
                                                   save=True) as arezzo_user_session:
        arezzo_user_session.favorites = sorted(set(arezzo_user_session.favorites) -
                                               {favorite})
    return ''


def _shut_up_flask():
    logger = logging.getLogger('werkzeug')
    logger.setLevel(logging.ERROR)
    app.logger.disabled = logger.disabled = True
    flask.cli.show_server_banner = lambda *args, **kwargs: None
    warnings.filterwarnings(
        'ignore',
        message="The 'environ['werkzeug.server.shutdown']' function is deprecated"
    )


class ArezzoServerThread(threading.Thread):
    def __init__(self, host: Optional[str] = DEFAULT_HOST,
                 port: Union[int, str, None] = DEFAULT_PORT, *,
                 quiet: bool = False) -> None:
        threading.Thread.__init__(self, daemon=True)
        self.host = host if (host is not None) else DEFAULT_HOST
        self.port = int(port) if (port is not None) else DEFAULT_PORT
        self.quiet = quiet

    def run(self) -> None:
        if self.quiet:
            _shut_up_flask()
        app.run(host=self.host, port=self.port)

    def shutdown(self) -> None:
        response = requests.post(urllib.parse.urljoin(DEFAULT_URL, 'shutdown'))
        response.raise_for_status()
        self.join()

    def __enter__(self) -> ArezzoServerThread:
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.shutdown()

    @property
    def url(self):
        host = '127.0.0.1' if self.host == '0.0.0.0' else self.host
        return make_url_from_host_port(host, self.port)





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import operator as operator_module
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar)
import collections.abc
import itertools

from .pointing import Point
from .. import int_crowding


class NoPoints(Exception):
    pass

class PointContainer(abc.ABC, collections.abc.Iterable, collections.abc.Container):

    intervals: tuple[int_crowding.Interval]

    def _arithmetic(self, other: ArithmeticOther, operator: Callable) -> IntCrowd:
        from .functions import intervals_to_points
        intervals = self._arithmetic_other_to_intervals(other)
        return IntCrowd(
            operator(set(intervals_to_points(self.intervals)), set(intervals_to_points(intervals))),
        )

    def __or__(self, other: ArithmeticOther) -> IntCrowd:
        intervals = self._arithmetic_other_to_intervals(other)
        return IntCrowd((self,) + intervals)

    def __and__(self, other: ArithmeticOther) -> IntCrowd:
        return self._arithmetic(other, operator_module.and_)

    def __xor__(self, other: ArithmeticOther) -> IntCrowd:
        return self._arithmetic(other, operator_module.xor)

    def __sub__(self, other: ArithmeticOther) -> IntCrowd:
        return self._arithmetic(other, operator_module.sub)

    def __contains__(self, other: ArithmeticOther) -> bool:
        from .functions import intervals_to_points
        intervals = self._arithmetic_other_to_intervals(other)
        return set(intervals_to_points(self.intervals)) >= set(intervals_to_points(intervals))

    def __bool__(self) -> bool:
        try:
            next(iter(self))
        except StopIteration:
            return False
        else:
            return True

    @abc.abstractmethod
    def __iter__(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def _reduce(self) -> tuple:
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError

    def __eq__(self, other: Any) -> bool:
        return (type(other) == type(self)) and (self._reduce() == other._reduce())

    def __hash__(self) -> int:
        return hash(self._reduce())

    @staticmethod
    def _arithmetic_other_to_intervals(arithmetic_other: ArithmeticOther) -> \
                                                                       tuple[int_crowding.Interval]:
        from .int_crowding import IntCrowd
        if isinstance(arithmetic_other, PointContainer):
            return arithmetic_other.intervals
        elif isinstance(arithmetic_other, Point):
            return (arithmetic_other.as_interval(),)
        elif isinstance(arithmetic_other, Iterable):
            intervals = []
            for item in arithmetic_other:
                if isinstance(item, PointContainer):
                    intervals.extend(item.intervals)
                else:
                    assert isinstance(item, Point)
                    intervals.append(item.as_interval())
            return tuple(intervals)

    def get_last_point(self):
        if not self:
            raise NoPoints
        last_interval = self.intervals[-1]
        return Point((item if isinstance(item, int) else item[1] - 1)
                     for item in last_interval.as_tuple())

    def trim_inclusive(self) -> IntCrowd:
        if self:
            assert self.dimensions >= 1
        return IntCrowd(int_crowding.Interval(interval.as_tuple()[:-1])
                        for interval in self.intervals)

    def get_single_point(self) -> Point:
        (point,) = self
        return point



ArithmeticOther = Union[PointContainer, Point, Iterable[Union[PointContainer, Point]], Any]

from .int_crowding import IntCrowd



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import pathlib
import logging
import string as string_module


SHOT_REWARD = -10
COLLISION_REWARD = -5
STARTING_REWARD = 0
NOTHING_REWARD = -1
FOOD_REWARD = 10

VISION_RANGE = 4
VISION_SIZE = VISION_RANGE * 2 + 1

LETTERS = string_module.ascii_uppercase

DEFAULT_BOARD_SIZE = 6 # 12
DEFAULT_N_FOOD_TILES = 5 # 20
DEFAULT_N_PLAYERS = 3 # 10



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import operator as operator_module
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar)
import collections.abc
import numbers
import itertools
import functools
import time as time_module

from marley import utils


def sleep_until(condition: Callable[[], Any], total_seconds: float = 10, *,
                step_seconds: float = 0.5, initial_seconds: float = 0,
                reset_condition: Optional[Callable[[], Any]] = None,
                exception: Union[Exception, Type[Exception]] = TimeoutError) -> None:
    if initial_seconds:
        time_module.sleep(initial_seconds)
    if condition():
        return
    div, mod = divmod(total_seconds - initial_seconds, step_seconds)
    n_steps = int(div + int(bool(mod)))
    while True:
        for _ in range(n_steps):
            time_module.sleep(step_seconds)
            if condition():
                return
            elif reset_condition is not None and reset_condition():
                break # to `while` loop, restarting the wait
        else:
            raise exception





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import functools
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict)

from .gaining import Gain, ThinGain, FatGain, SerialGain, ParallelGain, GainArg
from .int_crowding import IntCrowd


@functools.total_ordering
class JobType(abc.ABCMeta):
    def __get_long_name(self) -> str:
        return f'{self.__module__}.{self.__name__}'


    def __lt__(self, other: Union[JobType, Any]) -> bool:
        if not isinstance(other, JobType):
            raise NotImplementedError
        return self.__get_long_name() < other.__get_long_name()


@functools.total_ordering
class Job(metaclass=JobType):
    dimensions: int
    gain_type: Type[Gain]

    def get_parent_job_to_weight(self) -> Dict[Job, Weight]:
        return {}

    @abc.abstractmethod
    def _reduce(self) -> Tuple[Any]:
        pass

    def __repr__(self) -> str:
        return f'{type(self).__name__}{repr(self._reduce()[1:])}'

    def __hash__(self) -> int:
        return hash(self._reduce())

    def __eq__(self, other: Union[Any, Job]) -> bool:
        return (type(self) is type(other)) and self._reduce() == other._reduce()

    def create_gain(self, arg: GainArg = None) -> Gain:
        return self.gain_type(self, arg)

    def __lt__(self, other: Union[Job, Any]) -> bool:
        if not isinstance(other, Job):
            raise NotImplementedError
        return self._reduce() < other._reduce()




class ThinJob(Job):
    dimensions = 0
    gain_type = ThinGain

    def thin_sniff(self) -> Union[bool, ThinGain, None]:
        return None

    def thin_run(self) -> Optional[ThinGain]:
        return None

class FatJob(Job):
    max_chunk_size = 1
    dimensions: int

    def fat_sniff(self, fat_gain: FatGain) -> Union[FatGain, IntCrowd, None]:
        return None

    def fat_run(self, fat_gain: FatGain) -> Optional[FatGain]:
        return None


class SerialJob(FatJob):
    gain_type = SerialGain


class ParallelJob(FatJob):
    gain_type = ParallelGain


class JobSniffingJob(ThinJob):
    def __init__(self, target_gain: Gain):
        self.target_gain = target_gain

    def thin_run(self) -> Optional[Gain]:
        return self.target_gain.sniff()

    def thin_sniff(self) -> None:
        raise RuntimeError("Oh you didn't!")

    def _reduce(self) -> tuple:
        return (type(self), self.target_gain)




from .weighting import Weight, CalfWeight, TadpoleWeight



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import time
import multiprocessing.context
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict, Set)
from types import TracebackType
import concurrent.futures
import threading
import queue as queue_module
import atexit

import more_itertools

N_PRE_SPAWN_PROCESSES = 10

waiting_processes = queue_module.Queue()

spawning_thread = None

class SpawningThread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self, daemon=True)
        self.shutdown_requested = False
        self.start()

    def run(self):
        while not self.shutdown_requested:
            while waiting_processes.qsize() <= N_PRE_SPAWN_PROCESSES:
                process = PreSpawnProcess._create()
                waiting_processes.put(process)
                process.start()
            time.sleep(0.1)


class PreSpawnProcessType(type):

    def _create(cls):
        return type.__call__(cls)

    def __call__(cls, *, target: Callable, args: Iterable,
                 actually_create: bool = False) -> PreSpawnProcess:
        global spawning_thread
        if spawning_thread is None:
            spawning_thread = SpawningThread()
        process = waiting_processes.get()
        process.actually_start(target=target, args=args)
        return process


class PreSpawnProcess(multiprocessing.context.Process, metaclass=PreSpawnProcessType):
    def __init__(self, *args, **kwargs):
        multiprocessing.context.Process.__init__(self, *args, **kwargs)
        self.target_args_queue = multiprocessing.Queue()
        self.lock = multiprocessing.Lock()

    _is_started = False
    # def __init__(self, *, target: Callable, args: Iterable) -> None:
        # multiprocessing.context.Process.__init__(self, args=args)

    def run(self) -> None:
        try:
            target, args = self.target_args_queue.get(timeout=120)
        except queue_module.Empty:
            return
        else:
            target(*args)

    def start(self) -> None:
        with self.lock:
            if not self._is_started:
                multiprocessing.context.Process.start(self)
                self._is_started = True

    def actually_start(self, *, target: Callable, args: Iterable):
        self.target_args_queue.put((target, args))




class PreSpawnContext(multiprocessing.context.DefaultContext):
    Process = PreSpawnProcess

    def __init__(self):
        multiprocessing.context.DefaultContext.__init__(self, multiprocessing.get_context())


class PreSpawnProcessPoolExecutor(concurrent.futures.ProcessPoolExecutor):
    def __init__(self, max_workers=None, *, pre_spawn: bool = False) -> None:
        if pre_spawn:
            mp_context = PreSpawnContext()
        else:
            mp_context = None
        concurrent.futures.ProcessPoolExecutor.__init__(self, max_workers=max_workers,
                                                        mp_context=mp_context)



@atexit.register
def exit():
    if spawning_thread is not None:
        spawning_thread.shutdown_requested = True
    for process in more_itertools.iter_except(waiting_processes.get_nowait, queue_module.Empty):
        process.target_args_queue.put((sum, ((),)))
    time.sleep(0.1)
    for process in more_itertools.iter_except(waiting_processes.get_nowait, queue_module.Empty):
        process.target_args_queue.put((sum, ((),)))



if  __name__ ==  '__main__':
    pre_spawn_process = PreSpawnProcess(target=print, args=(7,))
    pre_spawn_process.start()
    pre_spawn_process.join()



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import operator as operator_module
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar)
import collections.abc
import itertools


from .pointing import Point
from .point_containing import PointContainer, ArithmeticOther
from .functions import canonicalize_intervals, intervals_to_points, points_to_intervals


class IntCrowd(PointContainer):

    def __init__(self, stuff: ArithmeticOther = ()) -> None:
        self.intervals = canonicalize_intervals(
            self._arithmetic_other_to_intervals(stuff)
        )
        all_dimensions = {interval.dimensions for interval in self.intervals}
        if not all_dimensions:
            self.dimensions = None
        else:
            (self.dimensions,) = all_dimensions

    def __iter__(self) -> Iterable[Point]:
        return itertools.chain.from_iterable(self.intervals)

    def _reduce(self) -> tuple:
        return (type(self), self.intervals)

    def __repr__(self) -> str:
        if not self:
            intervals_str = ''
        elif len(self.intervals) == 1:
            intervals_str = repr(self.intervals[0])
        else:
            assert len(self.intervals) >= 2
            intervals_str = repr(self.intervals)

        return f'{type(self).__name__}({intervals_str})'

    def to_savvy_content(self):
        return (self.intervals,)

    def separate(self, axes: Tuple[int]) -> Tuple[IntCrowd]:
        if not axes:
            return (self,)
        elif len(axes) > 1:
            raise NotImplementedError
        return tuple(IntCrowd(points) for (_, points) in
                     itertools.groupby(self, lambda point: tuple(point[axis] for axis in axes)))




# Copyright 2009-2017 Ram Rachum.
# This program is distributed under the MIT license.

'''
Defines the `WeakKeyIdentityDict` class.

See its documentation for more details.
'''
# todo: revamp

from __future__ import generator_stop

import weakref
import collections


__all__ = ['WeakKeyIdentityDict']


class IdentityRef(weakref.ref):
    '''A weak reference to an object, hashed by identity and not contents.'''

    def __init__(self, thing, callback=None):
        weakref.ref.__init__(self, thing, callback)
        self._hash = id(thing)


    def __hash__(self):
        return self._hash


class WeakKeyIdentityDict(collections.abc.MutableMapping):
    """
    A weak key dictionary which cares about the keys' identities.

    This is a fork of `weakref.WeakKeyDictionary`. Like in the original
    `WeakKeyDictionary`, the keys are referenced weakly, so if there are no
    more references to the key, it gets removed from this dict.

    The difference is that `WeakKeyIdentityDict` cares about the keys'
    identities and not their contents, so even unhashable objects like lists
    can be used as keys. The value will be tied to the object's identity and
    not its contents.
    """

    def __init__(self, dict_=None):
        self.data = {}
        def remove(k, selfref=weakref.ref(self)):
            self = selfref()
            if self is not None:
                del self.data[k]
        self._remove = remove
        if dict_ is not None: self.update(dict_)


    def __delitem__(self, key):
        del self.data[IdentityRef(key)]


    def __getitem__(self, key):
        return self.data[IdentityRef(key)]


    def __repr__(self):
        return f"<WeakKeyIdentityDict at {id(self)}>"


    def __setitem__(self, key, value):
        self.data[IdentityRef(key, self._remove)] = value


    def copy(self):
        """ D.copy() -> a shallow copy of D """
        new = WeakKeyIdentityDict()
        for key, value in self.data.items():
            o = key()
            if o is not None:
                new[o] = value
        return new


    def get(self, key, default=None):
        """ D.get(k[,d]) -> D[k] if k in D, else d.  d defaults to None. """
        return self.data.get(IdentityRef(key),default)


    def __contains__(self, key):
        try:
            wr = IdentityRef(key)
        except TypeError:
            return 0
        return wr in self.data


    has_key = __contains__


    def items(self):
        """ D.items() -> list of D's (key, value) pairs, as 2-tuples """
        L = []
        for key, value in list(self.data.items()):
            o = key()
            if o is not None:
                L.append((o, value))
        return L


    def iteritems(self):
        """ D.iteritems() -> an iterator over the (key, value) items of D """
        for wr, value in self.data.items():
            key = wr()
            if key is not None:
                yield key, value


    def iterkeyrefs(self):
        """Return an iterator that yields the weak references to the keys.

        The references are not guaranteed to be 'live' at the time
        they are used, so the result of calling the references needs
        to be checked before being used.  This can be used to avoid
        creating references that will cause the garbage collector to
        keep the keys around longer than needed.

        """
        return iter(self.data.keys())


    def iterkeys(self):
        """ D.iterkeys() -> an iterator over the keys of D """
        for wr in self.data.keys():
            obj = wr()
            if obj is not None:
                yield obj

    def __iter__(self):
        return iter(self.keys())


    def itervalues(self):
        """ D.itervalues() -> an iterator over the values of D """
        return iter(self.data.values())


    def keyrefs(self):
        """Return a list of weak references to the keys.

        The references are not guaranteed to be 'live' at the time
        they are used, so the result of calling the references needs
        to be checked before being used.  This can be used to avoid
        creating references that will cause the garbage collector to
        keep the keys around longer than needed.

        """
        return list(self.data.keys())


    def keys(self):
        """ D.keys() -> list of D's keys """
        L = []
        for wr in list(self.data.keys()):
            o = wr()
            if o is not None:
                L.append(o)
        return L


    def popitem(self):
        """ D.popitem() -> (k, v), remove and return some (key, value) pair
        as a 2-tuple; but raise KeyError if D is empty """
        while True:
            key, value = self.data.popitem()
            o = key()
            if o is not None:
                return o, value


    def pop(self, key, *args):
        """ D.pop(k[,d]) -> v, remove specified key and return the
        corresponding value. If key is not found, d is returned if given,
        otherwise KeyError is raised """
        return self.data.pop(IdentityRef(key), *args)


    def setdefault(self, key, default=None):
        """D.setdefault(k[,d]) -> D.get(k,d), also set D[k]=d if k not in D"""
        return self.data.setdefault(IdentityRef(key, self._remove),default)


    def update(self, dict=None, **kwargs):
        """ D.update(E, **F) -> None. Update D from E and F: for k in E: D[k] =
        E[k] (if E has keys else: for (k, v) in E: D[k] = v) then: for k in F:
        D[k] = F[k] """

        d = self.data
        if dict is not None:
            if not hasattr(dict, "items"):
                dict = type({})(dict)
            for key, value in dict.items():
                d[IdentityRef(key, self._remove)] = value
        if len(kwargs):
            self.update(kwargs)


    def __len__(self):
        return len(self.data)




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import os
import time
import threading
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict, Set)
from types import TracebackType
import concurrent.futures
from bidict import bidict
import uuid as uuid_module
import logging

import more_itertools

import marley.utils
from .weighting import Weight, CalfWeight, TadpoleWeight
from .gaining import ThinGain, FatGain, GainDyad
from .jobbing import Job, ThinJob, FatJob
from .antillesing import Antilles
from .cute_executing import CuteThreadExecutor, CuteProcessExecutor

logger = logging.getLogger(__name__)




class Supervisor:
    def __init__(self, *, use_multiprocessing: bool = True, n_workers: Optional[int],
                 sniff_only: bool = False, name: Optional[str] = None,
                 worker_setup_function: Optional[Callable] = None,
                 worker_setup_args: Sequence = ()) -> None:
        self.name = name or uuid_module.uuid4().hex[:6]
        self.antilles = Antilles(name=self.name)
        self.use_multiprocessing = use_multiprocessing
        self.n_workers = n_workers if n_workers is not None else os.cpu_count()
        self.requested_shut_down_when_finished_tasks = self.requested_shut_down = \
                                                                       self.main_loop_active = False
        self.future_to_pending_gain = bidict() # marley.utils.SelfLoggingBidict()
        self.pending_gain_to_future = self.future_to_pending_gain.inverse
        self.cute_executor = None
        self.sniff_only = sniff_only
        self.worker_setup_function = worker_setup_function
        self.worker_setup_args = worker_setup_args

    def get_executor_type(self) -> Type:
        return CuteProcessExecutor if self.use_multiprocessing else CuteThreadExecutor


    def main_loop(self) -> None:
        self.main_loop_active = True
        CuteExecutor = self.get_executor_type()
        with CuteExecutor(self.n_workers, worker_setup_function=self.worker_setup_function,
                          worker_setup_args=self.worker_setup_args) as self.cute_executor:
            while not self.requested_shut_down:
                doable_available_gains_iterator = self.antilles. \
                         iterate_doable_available_gains_and_mark_pending(sniff_only=self.sniff_only)
                while self.cute_executor.can_accept_more_futures():
                    logger.debug('Asking Antilles for a doable gain...')
                    try:
                        doable_available_gain = next(doable_available_gains_iterator)
                    except StopIteration:
                        logger.debug('Antilles is out of doable gains.')
                        break
                    else:
                        logger.debug(f'Antilles gave us this doable gain that we will now '
                                     f'schedule for execution: {doable_available_gain}')
                        future = self.cute_executor.submit(doable_available_gain)
                        future.gain = doable_available_gain # Todo: Maybe remove
                        assert doable_available_gain not in self.pending_gain_to_future
                        self.pending_gain_to_future[doable_available_gain] = future
                        logger.debug(f"Now that we scheduled that doable gain, there are "
                                     f"{len(self.pending_gain_to_future)} pending gains "
                                     f"that we're waiting for.")

                finished_gain_dyads = []
                for future in self.cute_executor.get_completed_futures(
                                                           wait_for_at_least_one_if_non_empty=True):
                    finished_gain_dyads.append(
                        GainDyad(
                            requested_gain=self.future_to_pending_gain.pop(future),
                            returned_gain=future.result()
                        )
                    )
                if finished_gain_dyads:
                    logger.debug(f'{len(finished_gain_dyads)} gains finished running, '
                                 f'reporting to Antilles.')
                    self.antilles.report_finished_gain_dyads(finished_gain_dyads)
                else:
                    if (self.requested_shut_down_when_finished_tasks and
                                                                    not self.cute_executor.futures):
                        break
                    else:
                        time.sleep(0.1) # That'll do, pig. That'll do.
        self.requested_shut_down_when_finished_tasks = self.requested_shut_down = \
                                                                       self.main_loop_active = False

    def request_shut_down(self, *, finish_jobs: bool = True) -> None:
        logger.debug(f'Asking supervisor to shut down, {finish_jobs=}')
        if self.main_loop_active:
            if finish_jobs:
                self.requested_shut_down_when_finished_tasks = True
            else:
                self.requested_shut_down = True


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import click
import webbrowser
import pathlib
import re
import time
from typing import (Optional, Tuple, Union, Container, Hashable, Iterator, Mapping,
                    Iterable, Any, Dict, FrozenSet, Callable, Type, Sequence, Set)
import logging
import platform
import sys

from marley import sharknado
import marley.commanding
import marley.sharknadoing
from .sharknadoing import *

logger = logging.getLogger(__name__)


@marley.commanding.marley_command_group.group(name='blackjack')
def command_group() -> None:
    pass

@click.option('-l/-n', '--last-project/--new-project', is_flag=True, default=True)
@click.option('-g', '--n-generations', default=3)
@click.option('-a', '--n-agents', default=2)
@click.option('--use-multiprocessing/--dont-use-multiprocessing', is_flag=True, default=True)
# @click.option('--n-evaluation-games', default=3_000)
@command_group.command()
def run(*, last_project: bool, n_generations: int, n_agents: int, use_multiprocessing: bool) -> None:
    if last_project:
        blackjack_project = BlackjackProject.get_last(marley.constants.swank_database)
    else:
        blackjack_project = BlackjackProject()
        blackjack_project.save()
    with marley.sharknadoing.MarleyShark(use_multiprocessing=use_multiprocessing) as shark:
        shark.add_directive_thin_jobs(blackjack_project.get_job(n_generations=n_generations,
                                                                n_agents=n_agents))


@click.option('-l/-n', '--last-project/--new-project', is_flag=True, default=True)
@click.option('--use-multiprocessing/--dont-use-multiprocessing', is_flag=True, default=True)
@command_group.command()
def write_chart(*, last_project: bool, use_multiprocessing: bool) -> None:
    if last_project:
        blackjack_project = BlackjackProject.get_last(marley.constants.swank_database)
    else:
        blackjack_project = BlackjackProject()
        blackjack_project.save()
    with marley.sharknadoing.MarleyShark(use_multiprocessing=use_multiprocessing) as shark:
        shark.add_directive_thin_jobs(blackjack_project.get_job(n_generations=0, n_agents=0))


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import math
import inspect
import re
import abc
import random
import itertools
import collections.abc
import statistics
import concurrent.futures
import enum
import functools
import numbers
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Mapping)
import dataclasses
import weakref

import more_itertools
import tensorflow as tf
import numpy as np

from marley import gamey
from .base import Observation, Action
from . import utils
from marley.jamswank.swanking import SimpleField, SavvyField, SwankField
from marley.gamey.gamey_swanking import GameySwank




class Policy(abc.ABC, GameySwank):
    '''
    Logic for deciding which action to take in a given observation.

    Your fancy machine-learning code goes here.
    '''

    def train(self, narratives: Sequence[gamey.Narrative]) -> Policy:
        # Override this method with your training code
        return self

    @abc.abstractmethod
    def get_next_action(self, observation: Observation) -> Action:
        raise NotImplementedError

    def __repr__(self) -> str:
        return f'{type(self).__name__}{self._extra_repr()}'

    def _extra_repr(self) -> str:
        return ('(<...>)' if inspect.signature(self.__init__).parameters else '()')


class SoloPolicy(Policy):
    @functools.cache
    def make_culture(self) -> aggregating.Culture:
        return aggregating.Culture.make_solo(self)


class SoloEpisodicPolicy(SoloPolicy):

    def get_score(self, make_initial_state: Callable[[], aggregating.State], n: int = 1_000) -> int:
        culture = self.make_culture()
        games = tuple(gaming.Game.from_state_culture(make_initial_state(), culture)
                      for _ in range(n))
        gaming.Game.multi_crunch(games)
        return np.mean(
            tuple(
                sum(payoff.get_single() for payoff in game.payoffs) for game in games
            )
        )


class RandomPolicy(Policy):
    def get_next_action(self, observation: Observation) -> Action:
        return random.choice(observation.legal_actions)


class QPolicy(Policy):
    '''A policy that calculates q-value for observation-actions.'''

    @property
    @functools.cache
    def q_map_cache(self) -> weakref.WeakKeyDictionary:
        return weakref.WeakKeyDictionary()


    @abc.abstractmethod
    def _get_qs_for_observations_uncached(self, observations: Sequence[Observation]) \
                                                            -> Tuple[Mapping[Action, numbers.Real]]:
        raise NotImplementedError

    def get_qs_for_observations(self, observations: Sequence[Observation]) \
                                                            -> Tuple[Mapping[Action, numbers.Real]]:
        result = [self.q_map_cache.get(observation, None) for observation in observations]
        uncached_indices = tuple(i for i, q_map in enumerate(result) if q_map is None)
        uncached_observations = tuple(observations[i] for i in uncached_indices)

        new_q_maps = self._get_qs_for_observations_uncached(uncached_observations)
        for i, observation, new_q_map in zip(uncached_indices, uncached_observations, new_q_maps):
            self.q_map_cache[observation] = new_q_map
            result[i] = new_q_map

        assert None not in result
        return tuple(result)

    def get_qs_for_observation(self, observation: Observation) -> Mapping[Action, numbers.Real]:
        return more_itertools.one(self.get_qs_for_observations((observation,)))


from . import aggregating
from . import gaming



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import re
import io
import abc
import itertools
import bisect
import pathlib
import collections.abc
from typing import Union, Iterable, Any, Mapping, TypeVar, Optional, Tuple, Sequence, Sized
import datetime as datetime_module
import string
import os
import random

import pyfakefs.fake_pathlib

from . import utils
from .jamming import (BaseWritableJamDatabase, Jam, JamKind, JamId, JamParchment, JamCorruption,
                      jam_id_pattern)


class JamFileDatabase(BaseWritableJamDatabase):
    def __init__(self, path: os.PathLike) -> None:
        self.path = pathlib.Path(path) if isinstance(path, str) else path

    @staticmethod
    def create_ethereal() -> JamFileDatabase:
        fake_filesystem = pyfakefs.fake_filesystem.FakeFilesystem()
        fake_pathlib_module = pyfakefs.fake_pathlib.FakePathlibModule(fake_filesystem)
        return JamFileDatabase(fake_pathlib_module.Path('/'))


    def _reduce(self) -> Tuple[Any]:
        return (self.path,)

    def _get_jam_kind_length(self, jam_kind_name: str) -> int:
        try:
            return utils.iterable_length((self.path / jam_kind_name).iterdir())
        except FileNotFoundError:
            return 0

    def _get_jam_kinds(self) -> Sequence[JamKind]:
        return tuple(self[jam_kind_path.name] for jam_kind_path in self.path.iterdir())

    def _get_jam_parchments(self, jam_kind_name: str) -> Sequence[JamParchment]:
        jam_kind = self[jam_kind_name]
        try:
            return tuple(
                jam_kind[jam_parchment_path.name] for jam_parchment_path in
                (self.path / jam_kind_name).iterdir()
                if jam_id_pattern.fullmatch(jam_parchment_path.name)
            )
        except FileNotFoundError:
            return ()

    def _get_jam_parchment_length(self, jam_kind_name: str, jam_id: Union[str, JamId]) -> int:
        (jam_id, jam_id_name) = JamId.parse(jam_id)
        path = self.path / jam_kind_name / jam_id_name
        try:
            return (path.stat().st_size) // jam_id.block_size
        except FileNotFoundError:
            return 0

    def _open_file_seeked(self, jam_kind_name: str, jam_id: JamId, *, start: int = 0,
                          at_end: bool = False, write: bool = False) -> io.IOBase:
        path = self.path / jam_kind_name / str(jam_id)
        if write:
            path.parent.mkdir(parents=True, exist_ok=True)
        if at_end:
            assert start == 0
            assert write
            file = path.open(f'ba')
            assert file.tell() % jam_id.block_size == 0
        else:
            try:
                file = path.open(f'b{"r+" if write else "r"}')
            except FileNotFoundError:
                path.touch()
                file = path.open(f'b{"r+" if write else "r"}')
            file.seek(jam_id.block_size * start)
        return file

    def _read_texts(self, jam_kind_name: str, jam_id: Union[str, JamId],
                    start: int, end: Optional[int]) -> Tuple[str]:
        (jam_id, _) = JamId.parse(jam_id)
        block_size = jam_id.block_size
        jam_parchment_length = self._get_jam_parchment_length(jam_kind_name, jam_id)
        if start < 0:
            start += jam_parchment_length
        if end is not None and end < 0:
            end += jam_parchment_length
        try:
            file = self._open_file_seeked(jam_kind_name, jam_id, start=start)
        except FileNotFoundError:
            return ()

        with file:
            result = []
            iterator = range(start, end) if (end is not None) else itertools.count(start)
            for _ in iterator:
                data = file.read(block_size)
                if len(data) == block_size:
                    result.append(data.decode('utf-8'))
                elif not data:
                    break
                else:
                    assert 0 < len(data) != block_size
                    raise JamCorruption
            return tuple(result)


    def _write_blobs(self, jam_kind_name: str, jam_id: Union[str, JamId],
                     blobs: Iterable[bytes], *, start: int = 0, append: bool = False) -> None:
        if append:
            assert start == 0
        (jam_id, _) = JamId.parse(jam_id)
        with self._open_file_seeked(jam_kind_name, jam_id, start=start, at_end=append,
                                    write=True) as file:
            for blob in blobs:
                assert len(blob) == jam_id.block_size
                file.write(blob)

    def __repr__(self) -> str:
        return f'{type(self).__name__}({repr(str(self.path))})'

    def __eq__(self, other: Any) -> bool:
        return (type(other) == type(self)) and (self._reduce() == other._reduce())

    def __hash__(self) -> int:
        return hash(self._reduce())



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import datetime as datetime_module
from typing import Iterable, Mapping, Union, Type, Optional, Any
import itertools
import functools
import filelock
import re
import collections.abc
import contextlib


from ..jamming import BaseJamDatabase, Jam, JamId, JamItem, JamKind, JamParchment, JamFileDatabase
from . import utils
from .exceptions import EmptyJam
from ..file_locking import FileLock




class SwankDatabase:
    def __init__(self, jam_database: BaseJamDatabase) -> None:
        self.jam_database = jam_database

    @classmethod
    def create_ethereal(cls) -> SwankDatabase:
        return cls(JamFileDatabase.create_ethereal())

    def get_jam_kind(self, arg: Union[Swank, Type[Swank], str]) -> JamKind:
        fixed_arg = type(arg) if isinstance(arg, Swank) else arg
        return self.jam_database[fixed_arg]

    def get_jam_item(self, swank: Swank) -> JamKind:
        assert swank.has_jam_id_and_index
        return self.get_jam_kind(swank)[swank.jam_id][swank.jam_index]

    def get_swank_type(self, arg: Union[Swank, Type[Swank], str]) -> JamKind:
        if isinstance(arg, str):
            jam_kind_name = arg
            return utils.name_to_type(jam_kind_name)
        else:
            if isinstance(arg, Swank):
                return type(arg)
            elif issubclass(arg, Swank):
                return arg


    def load_swank(self, swank_type_or_name: Union[Type[Swank], str], jam_id: Union[JamId, str],
                   jam_index: int) -> Swank:
        jam_kind = self.get_jam_kind(swank_type_or_name)
        swank_type = self.get_swank_type(swank_type_or_name)
        jam_item = jam_kind[jam_id][jam_index]
        jam = jam_item.read_jam()
        if jam is None:
            raise EmptyJam
        return swank_type._Swank__from_jam(jam, jam_id=jam_id, jam_index=jam_index,
                                           swank_database=self)

    def save_swank(self, swank: Swank) -> None:
        if not swank.has_jam_id_and_index:
            swank.jam_id = JamId.create(block_size=swank.default_block_size)
            swank.jam_index = 0
        jam_item = self.get_jam_item(swank)
        jam = swank._Swank__to_jam()
        jam_item.write_jam(jam)

    def iterate_latest(self, swank_type: Type[Swank]) -> Iterable[Swank]:
        return swank_type.iterate_latest(self)

    def _reduce(self) -> tuple:
        return (type(self), self.jam_database)

    def __eq__(self, other: Any) -> bool:
        return (type(other) == type(self)) and (self._reduce() == other._reduce())

    def __hash__(self) -> int:
        return hash(self._reduce())

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.jam_database})'




class SwankType(abc.ABCMeta):
    def __new__(mcls, name, bases, dict_) -> Type[Swank]:
        cls = abc.ABCMeta.__new__(mcls, name, bases, dict_)
        if len(cls.mro()) >= 3: # i.e. it's not the base `Swank` type, which doesn't have fields
            cls._Swank__fields = mcls.__get_fields(cls)
        return cls

    def __get_fields(cls) -> dict:
        fields = {}
        existing_names = set()
        for type_ in cls.mro():
            for name, value in vars(type_).items():
                if name in existing_names:
                    # This attribute was already overridden in a subclass, if it's a field we don't
                    # want to include it.
                    continue
                existing_names.add(name)
                if name.startswith('__'):
                    continue
                if isinstance(value, BaseField):
                    assert name not in fields
                    fields[name] = value
        return fields

    def iterate_latest(cls, swank_database: SwankDatabase) -> Iterable[Swank]:
        jam_parchments_by_latest = sorted(
            swank_database.get_jam_kind(cls),
            key=lambda jam_parcment: jam_parcment._get_path().stat().st_mtime,
            reverse=True,
        )
        for jam_parchment in jam_parchments_by_latest:
            for i in reversed(range(len(jam_parchment))):
                yield swank_database.load_swank(cls, jam_parchment.jam_id, i)

    def get_last(cls, swank_database: SwankDatabase) -> Swank:
        try:
            return next(cls.iterate_latest(swank_database))
        except StopIteration:
            raise IndexError

    def get_by_name(cls, swank_database: SwankDatabase, name_substring: str) -> Swank:
        for swank in cls.iterate_latest(swank_database):
            if name_substring in str(swank.jam_id):
                return swank
        raise LookupError



class Swank(metaclass=SwankType):
    default_block_size: int = 1_000
    jam_id: Optional[JamId] = None
    jam_index: Optional[int] = None

    def __init__(self, *, swank_database: SwankDatabase, jam_id: Optional[Union[JamId, str]] = None,
                 jam_index: Optional[int] = None, **kwargs):
        assert (jam_id, jam_index).count(None) in {0, 2}
        self.__field_values = {}
        self.jam_id = None if jam_id is None else JamId.cast(jam_id)
        self.jam_index = jam_index
        self.swank_database = swank_database
        assert set(kwargs) <= set(self._Swank__fields)

        for field_name, field in self.__fields.items():
            try:
                value = kwargs[field_name]
            except KeyError:
                value = field.get_default_value(swank_database)
            setattr(self, field_name, value)


    @property
    def has_jam_id_and_index(self):
        jam_id_exists = (self.jam_id is not None)
        jam_index_exists = (self.jam_index is not None)
        assert jam_id_exists == jam_index_exists
        return jam_id_exists


    @classmethod
    def __from_jam(cls, jam: Jam, *, jam_id: JamId, jam_index: int,
                   swank_database: SwankDatabase) -> Swank:
        fields = cls._Swank__fields
        swank = cls.__new__(cls)
        swank_cache = {(jam_id, jam_index): swank}

        ### Parsing jam into fields: ###############################################################
        #                                                                                          #
        kwargs = {}
        for full_field_name, value in jam.items():
            field_name, field_type_name = full_field_name.split('.')
            field = fields[field_name]
            assert field_type_name == field.field_type_name
            kwargs[field_name] = field.from_jam(value, swank_database=swank_database,
                                                swank_cache=swank_cache)
        #                                                                                          #
        ### Finished parsing jam into fields. ######################################################

        cls.__init__(swank, **kwargs,
                     jam_id=jam_id, jam_index=jam_index, swank_database=swank_database)
        return swank

    def __to_jam(self) -> Jam:
        fields = self._Swank__fields
        return {
            f'{name}.{field_type.field_type_name}':
                field_type.to_jam(getattr(self, name, None), self.swank_database)
            for name, field_type in fields.items()
        }

    @classmethod
    def load(cls, swank_database: SwankDatabase, jam_id: Union[JamId, str],
             jam_index: int) -> Swank:
        return swank_database.load_swank(cls, jam_id=jam_id, jam_index=jam_index)

    def reload(self) -> Swank:
        return type(self).load(self.swank_database, self.jam_id, self.jam_index)

    def save(self, *, all_parchment_fields: bool = True) -> Swank:
        from .fields import ParchmentField
        self.swank_database.save_swank(self)
        if all_parchment_fields:
            for field_name, field in self._Swank__fields.items():
                if isinstance(field, ParchmentField):
                    getattr(self, field_name).save()

    def _reduce(self) -> tuple:
        return (type(self), self.swank_database, self.jam_id, self.jam_index)

    def __eq__(self, other: Any) -> bool:
        return (
            (type(other) == type(self)) and
            self.has_jam_id_and_index and
            (self._reduce() == other._reduce())
        )

    def __hash__(self) -> int:
        return hash(self._reduce())

    def __repr__(self) -> str:
        text = ', '.join(
            f'{name}={value}' for name, value in self.__field_values.items()
        )
        return f'<{type(self).__name__}: {text}>'

    def to_savvy_content(self) -> tuple:
        return (str(self.jam_id), self.jam_index)

    @property
    @functools.cache
    def parchment_lock(self):
        jam_item = self.swank_database.get_jam_item(self)
        lock_path = jam_item.jam_parchment._get_lock_path()
        return FileLock(lock_path)



class SwankRef:
    def __init__(self, swank_database: SwankDatabase, swank_type: Type[Swank], jam_id: JamId,
                 jam_index: int) -> None:
        self.swank_database = swank_database
        self.swank_type = swank_type
        self.swank_type_name = utils.type_to_name(swank_type)
        self.jam_kind_name = utils.type_to_name(swank_type)
        assert isinstance(jam_id, JamId)
        self.jam_id = jam_id
        assert isinstance(jam_index, int)
        self.jam_index = jam_index

    @property
    def jam_item(self) -> JamItem:
        return self.swank_database.jam_database[self.jam_kind_name][self.jam_id][self.jam_index]

    def get(self) -> Swank:
        return self.swank_database.load_swank(self.jam_kind_name, self.jam_id, self.jam_index)

    @staticmethod
    def from_swank_or_ref(swank_or_ref: Optional[Union[Swank, SwankRef]]) -> Optional[SwankRef]:
        if swank_or_ref is None:
            return None
        elif isinstance(swank_or_ref, Swank):
            swank = swank_or_ref
            assert swank.has_jam_id_and_index
            return SwankRef(swank.swank_database, type(swank), swank.jam_id, swank.jam_index)
        else:
            assert isinstance(swank_or_ref, SwankRef)
            return swank_or_ref

    def _reduce(self) -> tuple:
        return (type(self), self.swank_database, self.swank_type, self.jam_id, self.jam_index)

    @property
    @functools.cache
    def parchment_lock(self):
        jam_kind = self.swank_database.get_jam_kind(self.swank_type)
        jam_item = jam_kind[self.jam_id][self.jam_index]
        lock_path = jam_item.jam_parchment._get_lock_path()
        return FileLock(lock_path)

    @contextlib.contextmanager
    def lock_and_load(self, *, save: bool = False) -> Swank:
        with self.parchment_lock:
            yield (swank := self.get())
            if save:
                swank.save()

    @contextlib.contextmanager
    def lock_and_load_or_create(self, *, save: bool = False) -> Swank:
        with self.parchment_lock:
            try:
                swank = self.get()
            except EmptyJam:
                swank = self.swank_type(swank_database=self.swank_database,
                                        jam_id=self.jam_id,
                                        jam_index=self.jam_index)
            yield swank
            if save:
                swank.save()

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: {self.swank_type_name}>'

    def __eq__(self, other: Any) -> bool:
        return (type(other) == type(self)) and (self._reduce() == other._reduce())

    def __hash__(self) -> int:
        return hash(self._reduce())








from .fields import BaseField


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations
from typing import Iterable, Tuple, Any

import sys
import itertools
import random
import enum
import functools
import numbers
import logging
from typing import Optional, Union

import numpy as np
import click
import more_itertools

from marley import gamey
from marley.jamswank.swanking import SimpleField, SavvyField, SwankDatabase
from marley.jamswank.jamming import JamId

logger = logging.getLogger(__name__)



def sum_cards(cards: Iterable[int]) -> int:
    result = 0
    n_aces = 0
    for card in cards:
        if 2 <= card <= 10:
            result += card
        else:
            assert card == 1
            n_aces += 1
            result += 11

    if result > 21: # Handling the aces:
        for _ in range(n_aces):
            result -= 10
            if result <= 21:
                break

    return result

assert sum_cards((1, 2, 3)) == 16
assert sum_cards((1, 2, 3, 5)) == 21
assert sum_cards((1, 2, 3, 6)) == 12


class _BlackjackActionType(type(gamey.Action), type(enum.Enum)):
    pass


class BlackjackAction(gamey.Action, enum.Enum, metaclass=_BlackjackActionType):
    hit = 'hit'
    stick = 'stick'
    wait = 'wait'

BlackjackAction.all_actions = (BlackjackAction.hit, BlackjackAction.stick,
                               BlackjackAction.wait)

_card_distribution = tuple(range(1, 10 + 1)) + (10,) * 3

def get_shuffled_deck() -> tuple:
    return tuple(gamey.utils.shuffled(_card_distribution * 4))


class BlackjackState(gamey.SoloState):
    player_cards = SimpleField()
    dealer_cards = SimpleField()
    deck = SimpleField()

    def __init__(self, *, player_cards: Tuple[int, ...], dealer_cards: Tuple[int, ...],
                 deck: Tuple[int, ...], player_id_to_observation: Optional[dict] = None,
                 jam_id: Optional[JamId] = None, jam_index: Optional[int] = None,
                 swank_database: Optional[SwankDatabase] = None) -> None:
        gamey.SoloState.__init__(self,
                                 player_cards=tuple(sorted(player_cards)),
                                 dealer_cards=tuple(sorted(dealer_cards)),
                                 deck=tuple(deck),
                                 player_id_to_observation=(player_id_to_observation or
                                                           {None: BlackjackObservation(self)}),
                                 jam_id=jam_id,
                                 jam_index=jam_index,
                                 swank_database=swank_database)

        self.player_stuck = (len(self.dealer_cards) >= 2)
        self.is_first_state = (len(self.player_cards) == len(self.dealer_cards) == 0)
        self.player_sum = sum_cards(self.player_cards)
        self.dealer_sum = sum_cards(self.dealer_cards)

        ### Calculating end value, if any: #########################################################
        #                                                                                          #
        if self.dealer_sum > 21:
            self.is_end = True
            self.reward = 1
        elif self.dealer_sum == 21:
            self.is_end = True
            assert self.player_sum <= 21
            self.reward = 0 if self.player_sum == 21 else -1
        elif 17 <= self.dealer_sum <= 20:
            assert self.player_stuck
            self.is_end = True
            if self.player_sum > self.dealer_sum:
                self.reward = 1
            elif self.player_sum < self.dealer_sum:
                self.reward = -1
            else:
                assert self.player_sum == self.dealer_sum
                self.reward = 0
        elif 2 <= self.dealer_sum <= 16:
            if self.player_stuck:
                self.is_end = False
                self.reward = 0
                assert self.player_sum <= 20
            else: # not self.player_stuck
                if self.player_sum > 21:
                    self.is_end = True
                    self.reward = -1
                elif self.player_sum == 21:
                    self.is_end = True
                    self.reward = 1
                else:
                    assert self.player_sum <= 20
                    self.is_end = False
                    self.reward = 0
        else:
            assert self.is_first_state
            self.is_end = False
            self.reward = 0

        #                                                                                          #
        ### Finished calculating end value, if any. ################################################

        if self.is_end:
            self.legal_actions = ()
        elif self.player_stuck:
            self.legal_actions = (BlackjackAction.wait,)
        elif self.is_first_state:
            self.legal_actions = (BlackjackAction.wait,)
        else:
            self.legal_actions = (BlackjackAction.hit, BlackjackAction.stick,)



    def get_next_reward_and_state(self, action: BlackjackAction) -> Tuple[numbers.Number,
                                                                          BlackjackState]:
        if action not in self.legal_actions:
            raise gamey.exceptions.IllegalAction(action)
        if self.is_first_state:
            state = BlackjackState(
                player_cards=self.deck[-2:],
                dealer_cards=(self.deck[-3],),
                deck=self.deck[:-3]
            )
        elif self.player_stuck or action == BlackjackAction.stick:
            state = BlackjackState(
                player_cards=self.player_cards,
                dealer_cards=(self.dealer_cards + self.deck[-1:]),
                deck=self.deck[:-1]
            )
        else:
            state = BlackjackState(
                player_cards=(self.player_cards + self.deck[-1:]),
                dealer_cards=self.dealer_cards,
                deck=self.deck[:-1]
            )
        return (state.reward, state)

    @staticmethod
    def make_initial() -> BlackjackState:
        return BlackjackState(player_cards=(), dealer_cards=(), deck=get_shuffled_deck())

    def __repr__(self) -> str:
        return (f'{type(self).__name__}'
                f'({self.player_cards}, {self.dealer_cards})')

    def _as_tuple(self) -> Tuple:
        return (self.player_cards, self.dealer_cards)

    def __hash__(self) -> int:
        return hash(self._as_tuple())

    def __eq__(self, other: Any) -> bool:
        return ((type(self) is type(other)) and
                (self._as_tuple() == other._as_tuple()))


class BlackjackObservation(gamey.Observation):
    def __init__(self, state: BlackjackState):
        self.state = state

    def to_savvy_content(self) -> tuple:
        return (self.state,)

    legal_actions = property(lambda self: self.state.legal_actions)
    player_sum = property(lambda self: self.state.player_sum)
    dealer_sum = property(lambda self: self.state.dealer_sum)
    player_cards = property(lambda self: self.state.player_cards)
    dealer_cards = property(lambda self: self.state.dealer_cards)
    player_stuck = property(lambda self: self.state.player_stuck)

    neural_dtype = np.dtype([('sequential_input', np.float64, 5)])

    def _to_neural(self) -> np.ndarray:
        sequential_input_array = np.array(
            tuple((
                self.player_sum / 21,
                1 in self.player_cards,
                self.dealer_sum / 21,
                1 in self.dealer_cards,
                float(self.player_stuck)
            ))
        )
        array = np.zeros((1,), dtype=self.neural_dtype)
        array['sequential_input'][0] = sequential_input_array
        return array




class BlackjackPolicy(gamey.SoloEpisodicPolicy):
    title: Optional[str] = None





class RandomPolicy(BlackjackPolicy, gamey.RandomPolicy):
    pass

class AlwaysHitPolicy(BlackjackPolicy):
    def get_next_action(self, observation: BlackjackState) -> BlackjackAction:
        return (BlackjackAction.hit if (BlackjackAction.hit in observation.legal_actions)
                else BlackjackAction.wait)

class AlwaysStickPolicy(BlackjackPolicy):
    '''A policy that always sticks, no matter what.'''
    def get_next_action(self, observation: BlackjackState) -> BlackjackAction:
        return (BlackjackAction.stick if (BlackjackAction.stick in observation.legal_actions)
                else BlackjackAction.wait)

    title = 'Always stick'

class ThresholdPolicy(BlackjackPolicy):
    '''
    A policy that sticks if the sum of cards is below the given threshold.
    '''
    def __init__(self, threshold: int = 17, *args, **kwargs) -> None:
        BlackjackPolicy.__init__(self, *args, **kwargs)
        self.threshold = threshold
        self.title = f'Stick at {threshold}+'


    def get_next_action(self, observation: BlackjackState) -> BlackjackAction:
        if BlackjackAction.wait in observation.legal_actions:
            return BlackjackAction.wait
        elif observation.player_sum >= self.threshold:
            return BlackjackAction.stick
        else:
            return BlackjackAction.hit

    def _extra_repr(self):
        return f'(threshold={self.threshold})'



class ModelFreeLearningPolicy(gamey.ModelFreeLearningPolicy, BlackjackPolicy):
    observation_type = BlackjackObservation
    action_type: Type[BlackjackAction] = SavvyField(lambda: BlackjackAction)
    default_block_size = 2_000_000








# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import pathlib
import logging.config
import datetime as datetime_module
import re
import shlex
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict)


from . import constants


MAX_LOG_FILES_TO_KEEP = 100

class NoisyThirdPartyFilter(logging.Filter):
    def filter(self, record):
        return record.name != 'filelock'

def clean_logs_folder() -> None:
    logs = tuple(constants.logs_folder.iterdir())
    if len(logs) >= MAX_LOG_FILES_TO_KEEP - 1:
        logs_to_delete = sorted(logs, key=lambda path: path.stat().st_ctime)[
                                                                 : -(MAX_LOG_FILES_TO_KEEP - 1)]
        for log_to_delete in logs_to_delete:
            log_to_delete.unlink()

log_file_path: Optional[pathlib.Path] = None
did_logging_setup: bool = False
is_verbose: Optional[bool] = None

def setup(*, verbose: bool = False, log_to_file: bool = True,
          existing_log_file_path: Optional[pathlib.Path] = None) -> None:
    global log_file_path, did_logging_setup, is_verbose
    if did_logging_setup:
        return
    is_verbose = verbose
    if log_to_file:
        if existing_log_file_path is not None:
            assert log_file_path is None
            log_file_path = existing_log_file_path
        else: # existing_log_file_path is None
            constants.logs_folder.mkdir(parents=True, exist_ok=True)
            clean_logs_folder()
            now = datetime_module.datetime.now()
            log_file_stem = re.sub('[^0-9]+', '-', now.isoformat(timespec='milliseconds'))
            assert re.fullmatch('[0-9-]+', log_file_stem)
            log_file_path = constants.logs_folder / f'{log_file_stem}.log'
            assert not log_file_path.exists()

        file_logging_setup = {
            'file': {
                'level': 'DEBUG',
                'class': 'logging.FileHandler',
                'filename': log_file_path,
                'mode': 'a',
                'formatter': 'verbose',
            },
        }

    else:
        assert existing_log_file_path is None
        assert log_file_path is None
        file_logging_setup = {}


    logging.config.dictConfig(
        {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'verbose': {
                    'format': '{levelname} {asctime} {module} p{process:d} t{thread:d} | {message}',
                    'style': '{',
                },
                'simple': {
                    'format': '{message}',
                    'style': '{',
                },
            },
            'filters': {
                'noisy_third_party_filter': {
                    '()': NoisyThirdPartyFilter,
                },
            },
            'handlers': {
                'console': {
                    'class': 'logging.StreamHandler',
                    'level': 'DEBUG' if verbose else 'INFO',
                    'formatter': 'simple',
                    'filters': ('noisy_third_party_filter',),
                },
                **file_logging_setup,
            },
            'root': {
                'handlers': ('console', *(('file',) if log_to_file else ())),
                'level': 'DEBUG',
            },
        }
    )

    logger = logging.getLogger(__name__)
    if log_to_file and not existing_log_file_path:
        logger.info(f'Log file: {shlex.quote(str(log_file_path))}')
    did_logging_setup = True


def get_logging_kwargs() -> dict:
    assert did_logging_setup
    return {
        'verbose': is_verbose,
        'log_to_file': (log_file_path is not None),
        'existing_log_file_path': log_file_path
    }




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import math
import inspect
import re
import abc
import random
import itertools
import collections.abc
import statistics
import concurrent.futures
import enum
import functools
import numbers
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar)
import dataclasses

import more_itertools
import numpy as np
import click
import logging

from marley import gamey
from marley.utils import ImmutableDict
from marley.jamswank.swanking import Swank, SimpleField, SavvyField, SwankDatabase
from . import utils
from . import exceptions
from .base import PlayerId, Action, Observation, Story
from .policing import Policy
from marley.gamey.gamey_swanking import GameySwank

logger = logging.getLogger(__name__)


def pluralize(s: str) -> str:
    known_values = {
        'policy': 'policies',
        'observation': 'observations',
        'reward': 'rewards',
        'mood': 'moods',
    }
    try:
        return known_values[s.lower()]
    except KeyError:
        return f'{s} objects'


class AggregateMixin(collections.abc.Mapping):
    _aggregate_value_type: Type


    @classmethod
    def __get_player_id_to_value_field_name(cls):
        try:
            return cls.__player_id_to_value_field_name
        except AttributeError:
            lowest_swank_base_type = min(
                (base_class for base_class in cls.mro() if issubclass(base_class, Swank) and
                 getattr(base_class, '_Swank__fields', None)),
                key=lambda base_class: len(base_class.mro())
            )
            ((cls.__player_id_to_value_field_name, field),) = \
                                                       lowest_swank_base_type._Swank__fields.items()
            assert isinstance(field, SavvyField)
            return cls.__player_id_to_value_field_name

    @property
    def __player_id_to_value(self) -> dict:
        return getattr(self, self.__get_player_id_to_value_field_name())

    @__player_id_to_value.setter
    def __player_id_to_value(self, d: dict) -> None:
        return setattr(self, self.__get_player_id_to_value_field_name(), d)

    def __getitem__(self, player_id: PlayerId) -> Any:
        return self.__player_id_to_value[player_id]

    def __iter__(self) -> Iterator:
        return iter(self.__player_id_to_value)

    def __len__(self) -> int:
        return len(self.__player_id_to_value)

    def __add__(self, other: AggregateMixin):
        if not isinstance(other, AggregateMixin):
            raise NotImplementedError
        return _CombinedAggregate((self, other))

    @classmethod
    def make_solo(cls, item: Union[numbers.Number, Action, Policy, Observation], /):
        return cls(**{cls.__get_player_id_to_value_field_name(): {None: item}})

    def get_single(self):
        return more_itertools.one(self.values())

    def __repr__(self):
        return (f'<{type(self).__name__} with {len(self)} '
                f'{pluralize(self._aggregate_value_type.__name__)}>')


class _CombinedAggregate(collections.abc.Mapping):
    def __init__(self, aggregates: Iterable[Union[AggregateMixin,
                                                  _CombinedAggregate]]) -> None:
        aggregates = tuple(aggregates)
        assert len(set(map(len, aggregates))) == 1 # They're all the same length.
        self.aggregates = []
        for aggregate in aggregates:
            if isinstance(aggregate, AggregateMixin):
                self.aggregates.append(aggregate)
            else:
                assert isinstance(aggregate, _CombinedAggregate)
                self.aggregates.extend(aggregate.aggregates)


    def __getitem__(self, player_id: PlayerId) -> Any:
        return tuple(aggregate[player_id] for aggregate in self.aggregates)

    def __iter__(self) -> Iterator:
        return iter(self.aggregates[0])

    def __len__(self) -> int:
        return len(self.aggregates[0])


class Activity(AggregateMixin, GameySwank):
    _aggregate_value_type = Action
    player_id_to_action = SavvyField(lambda: {})


class Payoff(AggregateMixin, GameySwank):
    _aggregate_value_type = numbers.Number
    player_id_to_reward = SavvyField(lambda: {})

    @staticmethod
    def make_zero(aggregate_player_value: AggregateMixin) -> Payoff:
        return Payoff(zip(aggregate_player_value, itertools.repeat(0)))



class Culture(AggregateMixin, GameySwank):
    _aggregate_value_type = Policy
    player_id_to_policy = SavvyField(lambda: {})

    def save(self) -> None:
        for policy in self.values():
            policy.save()
        GameySwank.save(self)


    def get_next_activity(self, state: State) -> Activity:
        return Activity(
            player_id_to_action={
                player_id: policy.get_next_action(observation) for
                player_id, (policy, observation) in (self + state).items()
            }
        )

    def train(self, make_initial_state: Callable[[], gamey.State], *, n_games: int = 1_000,
              max_game_length: Optional[int] = None, n_phases: int = 10) -> Policy:
        return more_itertools.last(
            self.train_iterate(
                make_initial_state, n_games=n_games, max_game_length=max_game_length,
                n_phases=n_phases
            )
        )


    def train_iterate(self, make_initial_state: Callable[[], gamey.State], *, n_games: int = 1_000,
                      max_game_length: Optional[int] = None, n_phases: Optional[int] = None,
                      log_average_reward: bool = True) -> Iterable[Policy]:
        from .gaming import Game
        culture = self
        games = []
        games_buffer_max_size = 10 * n_games
        for i_phase in range(n_phases):
            logger.debug(f'Starting training phase {i_phase} out of {n_phases or "infinity"}')
            new_games = tuple(Game.from_state_culture(make_initial_state(), culture)
                              for _ in range(n_games))
            Game.multi_crunch(new_games, n=max_game_length)

            if log_average_reward:
                average_reward = statistics.mean(
                    itertools.chain.from_iterable(new_game.total_payoff.values() for new_game
                                                  in new_games)
                )
                logger.debug(f'The average reward per agent per game is {average_reward:.2f}')

            games.extend(new_games)
            if len(games) >= games_buffer_max_size: # Truncate to avoid memory leaks
                del games[: -games_buffer_max_size]

            culture = type(self)(
                player_id_to_policy={
                    player_id: policy.train(tuple(game.narratives[player_id] for game in games))
                    for player_id, policy in culture.items()
                }
            )
            yield culture

    def train_iterate_progress_bar(self, make_initial_state: Callable[[], gamey.State], *,
                                   label: Optional[str] = None, n_games: int = 1_000,
                                   max_game_length: Optional[int] = None,
                                   n_phases: Optional[int] = None, **kwargs) -> Iterable[Policy]:
        if label is None:
            label = repr(self)
        logger.info(f'Training {label}...')
        progress_bar = click.progressbar(
            self.train_iterate(make_initial_state, n_games=n_games,
                               max_game_length=max_game_length, n_phases=n_phases),
            length=n_phases,
            **kwargs
        )
        with progress_bar:
            yield from progress_bar
        logger.info(f'Finished training {label}.')


    def train_progress_bar(self, make_initial_state: Callable[[], gamey.State], *,
                           label: Optional[str] = None, n_games: int = 1_000,
                           max_game_length: Optional[int] = None, n_phases: Optional[int] = None,
                           **kwargs) -> Iterable[Policy]:
        return more_itertools.last(
            self.train_iterate_progress_bar(
                make_initial_state=make_initial_state, label=label, n_games=n_games,
                max_game_length=max_game_length, n_phases=n_phases, **kwargs
            )
        )


class State(AggregateMixin, GameySwank):
    _aggregate_value_type = Observation
    player_id_to_observation = SavvyField(lambda: {})

    is_end: bool

    @staticmethod
    @abc.abstractmethod
    def make_initial(*args, **kwargs) -> State:
        '''Create an initial world state that we can start playing with.'''
        raise NotImplementedError

    @abc.abstractmethod
    def get_next_payoff_and_state(self, activity: Activity) -> Tuple[Payoff, State]:
        raise NotImplementedError

    @staticmethod
    def make_solo(solo_state: SoloState, /):
        assert isinstance(solo_state, SoloState)
        return solo_state


class SoloState(State):

    def get_next_payoff_and_state(self, activity: Activity) -> Tuple[Payoff, State]:
        reward, state = self.get_next_reward_and_state(more_itertools.one(activity.values()))
        payoff = Payoff.make_solo(reward)
        return (payoff, state)

    @abc.abstractmethod
    def get_next_reward_and_state(self, action: Action) -> Tuple[numbers.Number, SoloState]:
        raise NotImplementedError



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import os
import time as time_module
import threading
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict)
from types import TracebackType
import itertools
import uuid as uuid_module
import logging

from .weighting import Weight, CalfWeight, TadpoleWeight
from .gaining import ThinGain, FatGain
from .jobbing import Job, ThinJob, FatJob, JobSniffingJob
from .antillesing import Antilles
from .supervising import Supervisor
from .utils import sleep_until, ChangeTracker

logger = logging.getLogger(__name__)



class SharkThread(threading.Thread):
    def __init__(self, shark: Shark) -> None:
        threading.Thread.__init__(self, name=f'shark-{shark.name}')
        self.shark = shark

    def run(self) -> None:
        try:
            self.shark.supervisor.main_loop()
        except BaseException as base_exception:
            logger.exception('Shark thread exited because of an exception.')
            raise SystemExit(1) from base_exception


    def start(self) -> None:
        # This is making sure that the loop is marked as active before it starts, so if we try to
        # request a shutdown before it started, it still makes the request instead of thinking the
        # main loop isn't running.
        self.shark.supervisor.main_loop_active = True
        threading.Thread.start(self)



class Shark:
    def __init__(self, directive_thin_jobs: Union[ThinJob, Iterable[ThinJob]] = (), *,
                 name: Optional[str] = None, n_workers: Optional[int] = None,
                 use_multiprocessing: bool = True, start: bool = False,
                 sniff_only: bool = False, worker_setup_function: Optional[Callable] = None,
                 worker_setup_args: Sequence = ()) -> None:
        self.name = name or uuid_module.uuid4().hex[:6]
        self.shark_thread = None
        self.use_multiprocessing = use_multiprocessing
        self.n_workers = n_workers if n_workers is not None else os.cpu_count()
        self.supervisor = Supervisor(
            use_multiprocessing=use_multiprocessing, n_workers=self.n_workers, name=self.name,
            worker_setup_function=worker_setup_function, worker_setup_args=worker_setup_args
        )
        self.antilles = self.supervisor.antilles
        self.job_to_directive_gain = self.antilles.job_to_directive_gain
        self.add_directive_thin_jobs(directive_thin_jobs)
        self.sniff_only = sniff_only
        if start:
            self.start()

    @property
    def sniff_only(self) -> bool:
        return self._sniff_only

    @sniff_only.setter
    def sniff_only(self, value: bool) -> None:
        self.supervisor.sniff_only = self._sniff_only = value

    def start(self) -> None:
        assert self.shark_thread is None
        self.shark_thread = SharkThread(self)
        self.shark_thread.start()

    def add_directive_thin_jobs(self,
                                directive_thin_jobs: Union[ThinJob, Iterable[ThinJob]]) -> None:
        self.supervisor.antilles.job_to_directive_gain.add_thin_jobs(directive_thin_jobs)

    def shut_down(self, *, finish_jobs: bool = True) -> None:
        start_time = time_module.monotonic()
        self.supervisor.request_shut_down(finish_jobs=finish_jobs)
        logger.debug('Waiting for supervisor to finish and shark thread to stop...')
        self.shark_thread.join()
        end_time = time_module.monotonic()
        duration = end_time - start_time
        logger.debug(f'Shark thread stopped {duration:.2f}s after we asked it do. Shark shutdown '
                     f'is complete.')
        self.shark_thread = None

    def __enter__(self) -> Shark:
        assert not self.supervisor.requested_shut_down
        if self.shark_thread is None:
            self.start()
        return self

    def __exit__(self, exc_type: Type[Exception], exc_value: Exception,
                 exc_traceback: TracebackType) -> None:
        if self.shark_thread is not None:
            self.shut_down()

    def _sniffs_are_done(self) -> bool:
        return (
            not any(isinstance(job, JobSniffingJob) for job in
                    self.antilles.job_to_available_gain) and
            not self.antilles.job_to_sniff_pending_gain and
            not self.antilles.job_to_pending_gain and
            all((desired_gain in self.antilles.job_to_sniffed_gain[job])
                for job, desired_gain in self.antilles.job_to_desired_gain.items())
        )

    def wait_for_sniffs(self) -> None:
        # todo: This is shit, of course.
        assert self.sniff_only
        assert all(isinstance(job, JobSniffingJob) for job in self.antilles.job_to_pending_gain)
        logger.debug('Waiting for sniffs to finish...')
        change_tracker = ChangeTracker()
        sleep_until(self._sniffs_are_done, 200,
                    reset_condition=lambda: change_tracker(self.antilles._snapshot()))
        logger.debug(f'Done waiting for sniffs to finish. '
                     f'Current state:\n{self.antilles._snapshot().show()}')





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

import json
import math
import sys
import time
import itertools
import numbers
import collections.abc
import urllib.parse
from typing import Union, Optional, Iterator, Iterable, TypeVar, Callable, Tuple, TypeVar
import random

from browser import document, html, ajax, timer, window

from .utils import cool_ajax, add_parameters_to_url
from . import noice


jelly_database = noice.JellyDatabase('/jamming/')


def clamp(number, /, minimum, maximum):
    assert minimum <= maximum
    if number < minimum:
        return minimum
    elif number > maximum:
        return maximum
    else:
        return number



class Favorites:
    timepoint_text = property(lambda self: document.select('#timepoint-text')[0])
    timepoint_slider = property(lambda self: document.select('#timepoint-slider')[0])
    speed_slider = property(lambda self: document.select('#speed-slider')[0])
    play_pause_button = property(lambda self: document.select('button#play-pause')[0])
    update_button = property(lambda self: document.select('button#update-button')[0])
    keyboard_help_button = property(lambda self: document.select('button#keyboard-help-button')[0])
    play_pause_button_img = property(lambda self: document.select('button#play-pause img')[0])
    game_selector = property(lambda self: document.select('#game-selector')[0])
    loading_text = property(lambda self: document.select('#loading')[0])

favorites = Favorites()


_directions = ((0, -1), (1, 0), (0, 1), (-1, 0))
_directions_names = ('up', 'right', 'down', 'left')

def update_button_handler(event=None) -> None:
    update_game_names()
    timeline.transition_fetcher.update()

def focus_game_selector():
    favorites.game_selector.focus()


def keyboard_help_button_handler(event=None) -> None:
    window.alert(keyboard_help_text)


keyboard_help_text = '''\
P  Play / pause

G  Go to start of simulation
H  Go 5 states back
J  Go 1 state back

K  Go 1 state forward
L  Go 5 states forward
;  Go to end of simulation

I  Hold to slow down playback
O  Hold to speed up playback

S  Focus on game selector

?  Show this help screen
'''

Transition = TypeVar('Transition', bound=dict)

class Timeline(collections.abc.MutableSequence):

    def __init__(self) -> None:
        self._transitions = []
        self._needle = 0
        self._anchor_time = time.time()
        self._anchor_needle = self.needle
        self._anchor_speed = 3
        self._anchor_target = None
        self._anchor_play_after_target = None
        self.transition_fetcher = TransitionFetcher(self)
        self._game_name = None
        self._timer_id = None
        self.is_playing = False
        self.is_strong_playing = False
        self.speed = float(favorites.speed_slider.value)
        favorites.timepoint_slider.bind('input', self._on_timepoint_slider_change)
        favorites.speed_slider.bind('input', self._on_speed_slider_change)
        favorites.play_pause_button.bind('click', lambda event: self.toggle_playing())
        favorites.game_selector.bind(
            'change',
            lambda event: setattr(self, 'game_name',
                                  favorites.game_selector.selectedOptions[0].attrs['value'])
        )

    #######################################################################
    ### Defining sequence operations: #####################################
    #                                                                     #

    def __len__(self) -> int:
        return len(self._transitions)

    def __getitem__(self, i):
        return self._transitions[i]

    def __iter__(self) -> Iterator[Transition]:
        return iter(self._transitions)

    def __setitem__(self, i: Union[int, slice],
                    value: Union[Iterable[Transition], Transition]) -> None:
        raise NotImplementedError
        # self._transitions[i] = value

    def __delitem__(self, index: Union[int, slice]) -> None:
        del self._transitions[index]

    def insert(self, index: int, value: Transition) -> None:
        assert index == len(self._transitions) # Can only add items at the end
        self._transitions.insert(index, value)
        window.update_ui()

    def extend(self, values: Iterable[Transition]) -> None:
        for v in values:
            self.append(v)

    #                                                                     #
    ### Finished defining sequence operations. ############################
    #######################################################################


    @property
    def needle(self) -> float:
        return self._needle

    @needle.setter
    def needle(self, needle: float) -> None:
        self._needle = self._clamp_needle(needle)
        self._set_anchor()

    def _clamp_needle(self, needle):
        return clamp(round(needle), 0, max((len(self) - 1, 0)))


    def _on_timepoint_slider_change(self, event):
        self.needle = int(favorites.timepoint_slider.value)

    def _on_speed_slider_change(self, event):
        self.speed = float(favorites.speed_slider.value)
        self._set_anchor()

    def get_active(self) -> Transition:
        return self._transitions[math.ceil(self.needle)]

    def _set_anchor(self):
        self._anchor_time = time.time()
        self._anchor_needle = self.needle
        self._anchor_speed = self.speed
        self._anchor_target = False
        self._anchor_play_after_target = False


    def play(self, *, change_icon: bool = True) -> None:
        if not self.is_strong_playing:
            self._set_anchor()
            self.is_playing = self.is_strong_playing = True
            if change_icon:
                favorites.play_pause_button_img.attrs['src'] = 'pause.png'


    def pause(self, *, change_icon: bool = True) -> None:
        self.is_playing = self.is_strong_playing = False
        if change_icon:
            favorites.play_pause_button_img.attrs['src'] = 'play.png'
        self.needle = round(self.needle)

    def toggle_playing(self) -> None:
        if self.is_playing:
            self.pause()
        else:
            self.play()


    def skip(self, delta: numbers.Real) -> None:
        was_strong_playing = self.is_strong_playing

        target_needle = self._clamp_needle(self.needle + delta)

        distance_to_cover = target_needle - self.needle
        time_to_cover_distance = clamp(0.1 + ((abs(distance_to_cover) - 1) / 10),
                                       minimum=0.1, maximum=0.75)

        self._anchor_time = time.time()
        self._anchor_needle = self.needle
        self._anchor_speed = distance_to_cover / time_to_cover_distance
        self._anchor_target = target_needle
        self._anchor_play_after_target = was_strong_playing
        self.is_playing = True


    @property
    def game_name(self) -> str:
        return self._game_name

    @game_name.setter
    def game_name(self, new_game_name: str) -> None:
        self.pause()
        self._game_name = new_game_name
        self.clear()
        self._transitions.clear()
        self.needle = 0
        self.transition_fetcher.update()

    def change_speed(self, delta: numbers.Real) -> None:
        self.speed = clamp(self.speed + delta, 0.2, 8)
        window.update_ui()
        self._set_anchor()



class TransitionFetcher:
    delay = 20_000
    def __init__(self, timeline: Timeline) -> None:
        self.timeline = timeline
        self._timer_id = None
        self.game_name = None
        self._hid_loading_prompt = False

    def update(self):
        game_name = self.timeline.game_name
        if game_name != self.game_name:
            self.current_batch = 0
            self.game_name = game_name
        timer.clear_timeout(self._timer_id)
        self._do_next()

    def _do_next(self):
        def read_transitions(callback: Callable) -> None:
            start = len(self.timeline)
            end = start + 10
            jelly_database.read_jams('games', self.game_name, start, end, callback=callback)

        def callback(transitions: Tuple[dict]):
            self.timeline.extend(transitions)
            self._timer_id = timer.set_timeout(self._do_next, (100 if transitions else self.delay))
            if not self._hid_loading_prompt:
                favorites.loading_text.style.display = 'none' # On first update we hide it.
                self._hid_loading_prompt = True


        read_transitions(callback)

timeline = window.timeline = Timeline()

key_bindings = {
    'p': timeline.toggle_playing,

    'g': lambda: timeline.skip(- float('inf')),
    'h': lambda: timeline.skip(-5),
    'j': lambda: timeline.skip(-1),
    'k': lambda: timeline.skip(1),
    'l': lambda: timeline.skip(5),
    ';': lambda: timeline.skip(len(timeline)), #float('inf')),

    'i': lambda: timeline.change_speed(-0.2),
    'o': lambda: timeline.change_speed(0.2),

    'u': update_button_handler,
    's': focus_game_selector,

    '?': keyboard_help_button_handler,
}

def keypress_handler(event):
    key = event.key.lower()
    try:
        function = key_bindings[key]
    except KeyError:
        pass
    else:
        function()


def update_game_names(*, loop: bool = False, select_newest: bool = False) -> None:
    def _complete(game_names: Tuple[str]) -> None:
        for game_name in game_names:
            if not document.select(f'#game-selector option[value="{game_name}"]'):
                option = html.OPTION(value=game_name)
                option.textContent = game_name[:15]
                favorites.game_selector <= option
        if select_newest:
            latest_game_name = get_latest_game_name()
            latest_game_name_option = document.select(
                                      f'#game-selector option[value="{get_latest_game_name()}"]')[0]
            latest_game_name_option.selected = True
            timeline.game_name = latest_game_name
        if loop:
            timer.set_timeout(lambda: update_game_names(loop=True), 20_000)
    jelly_database.get_jam_parchment_names('games', _complete)

def get_latest_game_name():
    return max(option.attrs['value'] for option in document.select(f'#game-selector option'))


def main():
    document.bind('keypress', keypress_handler)
    update_game_names(loop=True, select_newest=True)
    window._animate_lock = False
    favorites.update_button.bind('click', update_button_handler)
    favorites.keyboard_help_button.bind('click', keyboard_help_button_handler)

    timer.request_animation_frame(window.animate)

    # jelly_database.get_jam_parchment_names('games', lambda x: print(f'Names: {x}'))
    # jam_id_name = '20210607-222217-587908-xwb96tt6tcfcwpx818qwyf6ynlznuy.000000001000'
    # jelly_database.read_texts('games', jam_id_name, 0, 1, callback=lambda x: print(x))
    # jelly_database.read_texts('games', jam_id_name, 0, 3, callback=lambda x: print(x))
    # jelly_database.read_texts('games', jam_id_name, 2, 17, callback=lambda x: print(x))
    # jelly_database.read_jams('games', jam_id_name, 0, 3, callback=lambda x: print(x))




main()


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

from typing import Optional

from .. import gamey


class GameyException(Exception):
    pass

class IllegalAction(GameyException):
    def __init__(self, action: Optional[gamey.Action] = None) -> None:
        self.action = action or "the given action"
        GameyException.__init__(self, f"Can't play {action} in this state.")


class GameOver(GameyException):
    '''You tried to go forward in a game that's already over.'''
    pass



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import itertools

from .pointing import Point
from .. import int_crowding


def intervals_to_points(intervals: Iterable[int_crowding.Interval]) -> Tuple[Point, ...]:
    return tuple(sorted(set(itertools.chain.from_iterable(intervals))))

def points_to_intervals(points: Iterable[Point]) -> Tuple[int_crowding.Interval, ...]:
    current_intervals = tuple(sorted(map(Point.as_interval, points)))
    if not current_intervals:
        return ()
    dimensions = current_intervals[0].dimensions
    def fluff():
        start_i = current_start.as_tuple()[i]
        end_i = current_current.as_tuple()[i]
        money_item = (start_i, end_i + 1) if start_i != end_i else start_i
        return int_crowding.Interval(current_current.as_tuple()[:i] + (money_item,) +
                                     current_current.as_tuple()[dimensions - d:])

    for d in range(dimensions):
        i = dimensions - d - 1
        new_intervals = []
        for _prefix, intervals in itertools.groupby(current_intervals,
                                                    lambda interval: interval.as_tuple()[:i]):
            current_start = current_current = next(intervals)

            for interval in intervals:
                if ((interval.as_tuple()[i] == current_current.as_tuple()[i] + 1) and
                    (interval.as_tuple()[i + 1:] == current_current.as_tuple()[i + 1:])):
                    current_current = interval
                else:
                    new_intervals.append(fluff())
                    current_start = current_current = interval
            new_intervals.append(fluff())

        current_intervals = tuple(new_intervals)

    return current_intervals

def canonicalize_intervals(intervals: Iterable[int_crowding.Interval] = ()) -> \
                                                                  Tuple[int_crowding.Interval, ...]:
    return points_to_intervals(intervals_to_points(intervals))





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import datetime as datetime_module
from typing import Iterable, Mapping, Union, Type, Optional, Any
import itertools
import collections.abc

from ...jamming import BaseJamDatabase, Jam, JamId, JamItem, JamKind, JamParchment
from ..core import Swank, SwankDatabase
from .base_field import BaseField


class SimpleField(BaseField):

    field_type_name = 'simple'

    def __init__(self, default_factory: Optional[Callable[[], Any]] = type(None)) -> None:
        self.default_factory = default_factory

    def from_jam(self, jam: Jam, swank_database: SwankDatabase, *, swank_cache: dict) -> Swank:
        return jam

    def to_jam(self, value: Optional[Any], swank_database: SwankDatabase) -> Jam:
        return value




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict)
import threading
import functools
import collections
import logging
import itertools
import random
import operator as operator_module
import uuid as uuid_module
import contextlib

import more_itertools

from .weighting import Weight, CalfWeight, TadpoleWeight
from .gaining import Gain, ThinGain, FatGain, GainArg, GainDyad
from .jobbing import Job, ThinJob, FatJob, JobSniffingJob
from .wedging import Wedge
from . import utils

import networkx as nx

logger = logging.getLogger(__name__)


class JobToGainDict(utils.CuteUserDict):
    def __setitem__(self, job: Job, gain: Gain) -> None:
        if gain:
            collections.UserDict.__setitem__(self, job, gain)
        else:
            try:
                del self[job]
            except KeyError:
                pass

class JobToItsGainDict(JobToGainDict):
    def __missing__(self, job: Job) -> Gain:
        return job.create_gain()

    def __setitem__(self, job: Job, gain_or_arg: Union[Gain, GainArg]) -> None:
        from .int_crowding import IntCrowd
        if isinstance(gain_or_arg, Gain):
            assert gain_or_arg.job == job
            gain = gain_or_arg
        else:
            assert isinstance(gain_or_arg, (IntCrowd, bool, type(None)))
            gain = job.create_gain(gain_or_arg)

        JobToGainDict.__setitem__(self, job, gain)



class JobToItsDirectiveGainDict(JobToItsGainDict):
    def __init__(self, antilles: Antilles, *args, **kwargs) -> None:
        self.antilles = antilles
        JobToItsGainDict.__init__(self, *args, **kwargs)


    def __setitem__(self, job: Job, gain: Gain) -> None:
        self.antilles.wedge.add_jobs(job)
        with self.antilles.lock, self.antilles._write_diff():
            if gain != self.get(job):
                JobToItsGainDict.__setitem__(self, job, gain)
                self.antilles._update_desired_and_available_gains_of_all_jobs()

    def __delitem__(self, job: Job) -> None:
        with self.antilles.lock, self.antilles._write_diff():
            gain = self.data.pop(job, None)
            if gain:
                self.antilles._update_desired_and_available_gains_of_all_jobs()

    def add_thin_jobs(self, thin_jobs: Union[ThinJob, Iterable[ThinJob]]) -> None:
        thin_jobs = (thin_jobs,) if isinstance(thin_jobs, ThinJob) else thin_jobs
        for thin_job in thin_jobs:
            self[thin_job] = True




class Antilles:
    def __init__(self, *, name: Optional[str] = None, keep_diffs: bool = True) -> None:
        self.name = name or uuid_module.uuid4().hex[:6]
        self.wedge = Wedge()
        self.lock = threading.RLock()

        self.keep_diffs = keep_diffs
        self.diffs = [] if keep_diffs else None

        # Gains that were run are called "finished":
        self.job_to_finished_gain = JobToItsGainDict()

        # Gains that the supervisor is working on are called "pending":
        self.job_to_pending_gain = JobToItsGainDict()

        # Gains that were finished in a previous run:
        self.job_to_preexisting_gain = JobToItsGainDict()

        # Gains that were sniffed, i.e. we checked for preexisting gain:
        self.job_to_sniffed_gain = JobToItsGainDict()

        # Gains that the supervisor was given to sniff:
        self.job_to_sniff_pending_gain = JobToItsGainDict()

        # Gains that we were explicitly asked by the user to run are called "directive":
        self.job_to_directive_gain = JobToItsDirectiveGainDict(self)

        # Gains that we want to be finished are called "desired", even if they're already finished:
        self.job_to_desired_gain = JobToItsGainDict()
        # A gain is desired either if it's directive, or if it's a dependency of a directive gain.

        # Gains that are desired, were not given to the supervisor, and are not blocked by any
        # dependencies are called "available":
        self.job_to_available_gain = JobToItsGainDict()

        self.job_to_child_job_to_desired_gain_from_it = collections.defaultdict(JobToGainDict)

        self.doable_available_gain_log = [] # todo remove?


    # def _update_desired_and_available_gain_from_directive_and_children(self, job: Job) -> None:
        # self.job_to_desired_gain[job] = utils.union((self.job_to_directive_gain[job],) + tuple(
                                     # self.job_to_child_job_to_desired_gain_from_it[job].values()))
        # self._update_available_gains_for_jobs(job)


    def _update_desired_and_available_gains_of_all_jobs(self) -> None:
        # Inefficient implementation.
        with self.lock:
            job_to_desired_gain = JobToItsGainDict()
            job_to_child_job_to_desired_gain_from_it = collections.defaultdict(JobToGainDict)

            all_jobs = set(self.wedge.iterate_predecessor_jobs(self.job_to_directive_gain,
                                                               include_self=True))
            dirty_jobs = set(all_jobs)
            while dirty_jobs:
                job = random.choice(tuple(dirty_jobs))
                dirty_jobs.remove(job)
                old_guess_for_desired_gain = job_to_desired_gain[job]
                current_guess_for_desired_gain = (
                    utils.union((self.job_to_directive_gain[job],) + tuple(
                    job_to_child_job_to_desired_gain_from_it[job].values()))
                ).get_gain_desired_by_self()
                if current_guess_for_desired_gain != old_guess_for_desired_gain:
                    job_to_desired_gain[job] = current_guess_for_desired_gain
                    sniffed_not_preexisting_part_of_current_guess_for_desired_gain = (
                        current_guess_for_desired_gain & self.job_to_sniffed_gain[job] -
                        self.job_to_preexisting_gain[job]
                    )
                    dirty_jobs.update(job.get_parent_job_to_weight())

                    for parent_job, weight in job.get_parent_job_to_weight().items():
                        weight: Weight
                        desired_gain_from_child = weight.get_desired_parent_gain(
                            sniffed_not_preexisting_part_of_current_guess_for_desired_gain,
                            parent_job
                        )
                        job_to_child_job_to_desired_gain_from_it[parent_job][job] = \
                                                                             desired_gain_from_child

            self.job_to_desired_gain = job_to_desired_gain
            self.job_to_child_job_to_desired_gain_from_it = job_to_child_job_to_desired_gain_from_it

            self._update_available_gains(all_jobs)

            for job in (set(self.job_to_available_gain) - set(all_jobs)):
                if not isinstance(job, JobSniffingJob):
                    del self.job_to_available_gain[job]




    def report_finished_gain_dyads(self, gain_dyads: Iterable[GainDyad]) -> None:
        with self.lock, self._write_diff():
            jobs_that_have_new_finished_gains = set()
            need_to_update_desired_gains_of_all_jobs = False
            for gain_dyad in gain_dyads:
                gain_dyad: GainDyad
                # print(gain_dyad)
                requested_gain = gain_dyad.requested_gain
                returned_gain = gain_dyad.returned_gain
                assert requested_gain in self.job_to_pending_gain[requested_gain.job]
                assert requested_gain not in self.job_to_finished_gain[requested_gain.job]
                if isinstance(requested_gain.job, JobSniffingJob):
                    need_to_update_desired_gains_of_all_jobs = True
                    job_sniffing_job = requested_gain.job
                    target_gain = job_sniffing_job.target_gain
                    assert target_gain in self.job_to_sniff_pending_gain[target_gain.job]
                    assert target_gain not in self.job_to_sniffed_gain[target_gain.job]
                    self.job_to_sniff_pending_gain[target_gain.job] -= target_gain
                    self.job_to_sniffed_gain[target_gain.job] |= target_gain
                    self.job_to_pending_gain[requested_gain.job] -= requested_gain
                    self.job_to_finished_gain[requested_gain.job] |= requested_gain
                    if returned_gain:
                        assert returned_gain.job == target_gain.job
                        self.job_to_finished_gain[returned_gain.job] |= returned_gain
                        self.job_to_preexisting_gain[returned_gain.job] |= returned_gain
                        jobs_that_have_new_finished_gains.add(returned_gain.job)
                else:
                    finished_gain = (gain_dyad.requested_gain if gain_dyad.returned_gain is None
                                     else gain_dyad.returned_gain)
                    self.job_to_pending_gain[requested_gain.job] -= finished_gain
                    self.job_to_finished_gain[requested_gain.job] |= finished_gain
                    jobs_that_have_new_finished_gains.add(requested_gain.job)


            if need_to_update_desired_gains_of_all_jobs:
                self._update_desired_and_available_gains_of_all_jobs()
            else:
                directly_affected_jobs = self.wedge.get_child_jobs(
                    jobs_that_have_new_finished_gains,
                    include_self=True
                )
                self._update_available_gains(directly_affected_jobs)

            for job in (set(self.job_to_available_gain) | set(self.job_to_finished_gain)):
                assert not self.job_to_available_gain[job] & self.job_to_finished_gain[job]

    def _update_available_gains(self, jobs: Union[Job, Iterable[Job]]) -> None:
        # todo: The more efficient design would involve getting the new gains for the parents, so
        # we could calculate only the delta instead of the entire thing. But we can hold off on that
        # for now.
        with self.lock:
            jobs: set = {jobs} if isinstance(jobs, Job) else set(jobs)
            for job in jobs:
                job: Job
                desired_gain: Gain = self.job_to_desired_gain[job]
                desired_unsniffed_not_sniff_pending_gain = (
                    desired_gain - self.job_to_sniffed_gain[job] -
                    self.job_to_sniff_pending_gain[job]
                )
                if desired_unsniffed_not_sniff_pending_gain:
                    self.job_to_available_gain[
                                    JobSniffingJob(desired_unsniffed_not_sniff_pending_gain)] = True
                desired_sniffed_not_preexisting_gain = (
                    desired_gain & self.job_to_sniffed_gain[job] -
                    self.job_to_preexisting_gain[job]
                )
                unblocked_gain_components = [
                    desired_sniffed_not_preexisting_gain.
                                                    get_gain_unblocked_by_finished_gain_of_same_job(
                                                            self.job_to_finished_gain[job]
                                                        )
                ]
                for parent_job, weight in job.get_parent_job_to_weight().items():
                    unblocked_gain_components.append(
                        desired_sniffed_not_preexisting_gain.
                                                         get_gain_unblocked_by_finished_parent_gain(
                            self.job_to_finished_gain[parent_job],
                            weight
                        )
                    )

                self.job_to_available_gain[job] = (utils.intersection(unblocked_gain_components) -
                                                   self.job_to_finished_gain[job] -
                                                   self.job_to_pending_gain[job])

    def iterate_doable_available_gains_and_mark_pending(self, *,
                                                        sniff_only: bool = False) -> Iterator[Gain]:
        while True:
            with self.lock:
                if sniff_only:
                    job_pool = tuple(job for job in self.job_to_available_gain
                                     if isinstance(job, JobSniffingJob))
                else:
                    job_pool = tuple(self.job_to_available_gain)
                if not job_pool:
                    break
                job = random.choice(job_pool)
                available_gain = self.job_to_available_gain[job]
                if not available_gain:
                    continue
                doable_available_gain = available_gain.get_doable()
                assert doable_available_gain not in self.doable_available_gain_log
                with self._write_diff():
                    self.doable_available_gain_log.append(doable_available_gain)
                    self.job_to_available_gain[job] -= doable_available_gain
                    self.job_to_pending_gain[job] |= doable_available_gain
                    if isinstance(doable_available_gain.job, JobSniffingJob):
                        target_gain = doable_available_gain.job.target_gain
                        self.job_to_sniff_pending_gain[target_gain.job] |= target_gain
            yield doable_available_gain


    @contextlib.contextmanager
    def _write_diff(self) -> AntillesDiffRecorder:
        if not self.keep_diffs:
            yield
        else:
            with AntillesDiffRecorder(self) as antilles_change_recorder:
                yield
            if antilles_change_recorder.antilles_diff:
                self.diffs.append(antilles_change_recorder.antilles_diff)
                logger.debug(f'Antilles diff:\n{antilles_change_recorder.antilles_diff.show()}')

    def _snapshot(self) -> AntillesSnapshot:
        return AntillesSnapshot.create_from_antilles(self)


    def diffs_as_dict(self) -> dict:
        assert self.keep_diffs
        return tuple(antilles_diff.show() for antilles_diff in self.diffs)

    def show_diffs(self) -> str:
        assert self.keep_diffs
        return '\n'.join(antilles_diff.show() for antilles_diff in self.diffs)


from .antilles_diffing import AntillesDiffRecorder, AntillesSnapshot


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

from typing import Optional

from .core import *
from .fields import *
from .singletoning import SingletonSwank




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import re
import io
import abc
import itertools
import bisect
import pathlib
import collections.abc
from typing import (Union, Iterable, Any, Mapping, TypeVar, Optional, Tuple, Sequence, Sized,
                    Callable)
import datetime as datetime_module
import string
import random
import flask
import json

from . import utils
from . import jamming

routes = {}
flask_app_name_to_jam_database = {}


def add_to_routes(path: str) -> Callable:
    def inner(view_function: Callable) -> Callable:
        assert path.startswith('/')
        assert path not in routes
        routes[path] = view_function
        return view_function
    return inner

def _get_jam_database_for_current_app() -> jamming.BaseJamDatabase:
    return flask_app_name_to_jam_database[flask.current_app.name]

@add_to_routes('/')
def jamming_index() -> str:
    return 'Nothing here'

@add_to_routes('/<jam_kind_name>/')
def jamming_get_jam_parchment_names(jam_kind_name: str) -> str:
    assert re.fullmatch('[a-z_]+', jam_kind_name)
    jam_database = _get_jam_database_for_current_app()
    return json.dumps(tuple(str(jam_parchment.jam_id) for jam_parchment in
                            jam_database[jam_kind_name]))



@add_to_routes('/<jam_kind_name>/<jam_id_name>/<int:start>..<int:end>/')
def jamming_read_texts(jam_kind_name: str, *, jam_id_name: str, start: int,
                       end: Optional[int]) -> str:
    jam_database = _get_jam_database_for_current_app()
    assert re.fullmatch('^[A-Za-z_.]+$', jam_kind_name)
    jam_id = jamming.JamId(jam_id_name)
    jam_parchment: jamming.JamParchment = jam_database[jam_kind_name][jam_id]
    return ''.join(jam_parchment.read_texts(start, end))


@add_to_routes('/<jam_kind_name>/<jam_id_name>/')
def jamming_get_parchment_info(jam_kind_name: str, *, jam_id_name: str) -> str:
    jam_database = _get_jam_database_for_current_app()
    assert re.fullmatch('^[A-Za-z_.]+$', jam_kind_name)
    jam_id = jamming.JamId(jam_id_name)
    jam_parchment: jamming.JamParchment = jam_database[jam_kind_name][jam_id]
    return json.dumps({'length': len(jam_parchment)})


def add_to_flask_app(jam_database: jamming.BaseJamDatabase, app: flask.Flask,
                     prefix: str = '/api/jam') -> None:
    assert not prefix.endswith('/')
    assert app.name not in flask_app_name_to_jam_database
    flask_app_name_to_jam_database[app.name] = jam_database
    assert re.fullmatch('^/[a-z0-9_/]+$', prefix)
    for path, view_function in routes.items():
        full_path = utils.removesuffix(prefix + path, '/')
        app.add_url_rule(full_path, view_func=view_function)





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

'''A collection of general-purpose tools.'''

from __future__ import annotations

import tempfile
import shutil
import io
import sys
import pathlib
import contextlib
import builtins
import time as time_module
import threading
import copyreg
import contextlib
from typing import Iterable, Iterator, Hashable
import time
import datetime as datetime_module


def iterate_deduplicated(iterable: Iterable[Hashable], seen: Iterable[Hashable] = ()) \
                                                                              -> Iterator[Hashable]:
    seen = set(seen)
    for item in iterable:
        if item in seen:
            continue
        else:
            yield item
            seen.add(item)


def pickle_lock(lock):
    return (threading.Lock, ())

copyreg.pickle(type(threading.Lock()), pickle_lock)


def pickle_r_lock(r_lock):
    return (threading.RLock, ())

copyreg.pickle(type(threading.RLock()), pickle_r_lock)


# def pickle_stack_summary(stack_summary):
    # return (tensorflow.compat.v1.flags.tf_decorator.tf_stack.StackSummary, ())


# copyreg.pickle(tensorflow.compat.v1.flags.tf_decorator.tf_stack.StackSummary,
               # pickle_stack_summary)



@contextlib.contextmanager
def BlankContextManager():
    yield

class NotInDict:
    '''Object signifying that the key was not found in the dict.'''


class TempValueSetter(object):
    '''
    Context manager for temporarily setting a value to a variable.

    The value is set to the variable before the suite starts, and gets reset
    back to the old value after the suite finishes.
    '''

    def __init__(self, variable, value, assert_no_fiddling=True):
        '''
        Construct the `TempValueSetter`.

        `variable` may be either an `(object, attribute_string)`, a `(dict,
        key)` pair, or a `(getter, setter)` pair.

        `value` is the temporary value to set to the variable.
        '''

        self.assert_no_fiddling = assert_no_fiddling


        #######################################################################
        # We let the user input either an `(object, attribute_string)`, a
        # `(dict, key)` pair, or a `(getter, setter)` pair. So now it's our job
        # to inspect `variable` and figure out which one of these options the
        # user chose, and then obtain from that a `(getter, setter)` pair that
        # we could use.

        bad_input_exception = Exception(
            '`variable` must be either an `(object, attribute_string)` pair, '
            'a `(dict, key)` pair, or a `(getter, setter)` pair.'
        )

        try:
            first, second = variable
        except Exception:
            raise bad_input_exception
        if hasattr(first, '__getitem__') and hasattr(first, 'get') and \
           hasattr(first, '__setitem__') and hasattr(first, '__delitem__'):
            # `first` is a dictoid; so we were probably handed a `(dict, key)`
            # pair.
            self.getter = lambda: first.get(second, NotInDict)
            self.setter = lambda value: (first.__setitem__(second, value) if
                                         value is not NotInDict else
                                         first.__delitem__(second))
            ### Finished handling the `(dict, key)` case. ###

        elif callable(second):
            # `second` is a callable; so we were probably handed a `(getter,
            # setter)` pair.
            if not callable(first):
                raise bad_input_exception
            self.getter, self.setter = first, second
            ### Finished handling the `(getter, setter)` case. ###
        else:
            # All that's left is the `(object, attribute_string)` case.
            if not isinstance(second, str):
                raise bad_input_exception

            parent, attribute_name = first, second
            self.getter = lambda: getattr(parent, attribute_name)
            self.setter = lambda value: setattr(parent, attribute_name, value)
            ### Finished handling the `(object, attribute_string)` case. ###

        #
        #
        ### Finished obtaining a `(getter, setter)` pair from `variable`. #####


        self.getter = self.getter
        '''Getter for getting the current value of the variable.'''

        self.setter = self.setter
        '''Setter for Setting the the variable's value.'''

        self.value = value
        '''The value to temporarily set to the variable.'''

        self.active = False


    def __enter__(self):

        self.active = True

        self.old_value = self.getter()
        '''The old value of the variable, before entering the suite.'''

        self.setter(self.value)

        # In `__exit__` we'll want to check if anyone changed the value of the
        # variable in the suite, which is unallowed. But we can't compare to
        # `.value`, because sometimes when you set a value to a variable, some
        # mechanism modifies that value for various reasons, resulting in a
        # supposedly equivalent, but not identical, value. For example this
        # happens when you set the current working directory on Mac OS.
        #
        # So here we record the value right after setting, and after any
        # possible processing the system did to it:
        self._value_right_after_setting = self.getter()

        return self


    def __exit__(self, exc_type, exc_value, exc_traceback):

        if self.assert_no_fiddling:
            # Asserting no-one inside the suite changed our variable:
            assert self.getter() == self._value_right_after_setting

        self.setter(self.old_value)

        self.active = False

class OutputCapturer(object):
    '''
    Context manager for catching all system output generated during suite.

    Example:

        with OutputCapturer() as output_capturer:
            print('woo!')

        assert output_capturer.output == 'woo!\n'

    The boolean arguments `stdout` and `stderr` determine, respectively,
    whether the standard-output and the standard-error streams will be
    captured.
    '''
    def __init__(self, stdout=True, stderr=True):
        self.string_io = io.StringIO()

        if stdout:
            self._stdout_temp_setter = \
                TempValueSetter((sys, 'stdout'), self.string_io)
        else: # not stdout
            self._stdout_temp_setter = BlankContextManager()

        if stderr:
            self._stderr_temp_setter = \
                TempValueSetter((sys, 'stderr'), self.string_io)
        else: # not stderr
            self._stderr_temp_setter = BlankContextManager()

    def __enter__(self):
        '''Manage the `OutputCapturer`'s context.'''
        self._stdout_temp_setter.__enter__()
        self._stderr_temp_setter.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        # Not doing exception swallowing anywhere here.
        self._stderr_temp_setter.__exit__(exc_type, exc_value, exc_traceback)
        self._stdout_temp_setter.__exit__(exc_type, exc_value, exc_traceback)

    output = property(lambda self: self.string_io.getvalue(),
                      doc='''The string of output that was captured.''')


class TempSysPathAdder(object):
    '''
    Context manager for temporarily adding paths to `sys.path`.

    Removes the path(s) after suite.

    Example:

        with TempSysPathAdder('path/to/fubar/package'):
            import fubar
            fubar.do_stuff()

    '''
    def __init__(self, addition):
        self.addition = [str(addition)]


    def __enter__(self):
        self.entries_not_in_sys_path = [entry for entry in self.addition if
                                        entry not in sys.path]
        sys.path += self.entries_not_in_sys_path
        return self


    def __exit__(self, *args, **kwargs):

        for entry in self.entries_not_in_sys_path:

            # We don't allow anyone to remove it except for us:
            assert entry in sys.path

            sys.path.remove(entry)


def _zip_strict(*iterables):
    if not iterables:
        return
    iterators = tuple(iter(iterable) for iterable in iterables)
    try:
        while True:
            items = []
            for iterator in iterators:
                items.append(next(iterator))
            yield tuple(items)
    except StopIteration:
        if items:
            raise ValueError
        sentinel = object()
        for _, iterator in enumerate(iterators[1:], 1):
            if next(iterator, sentinel) is not sentinel:
                raise ValueError

def zip(*iterables, strict: bool = False):
    if strict is False:
        return builtins.zip(*iterables)
    else:
        return _zip_strict(*iterables)





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import datetime as datetime_module
from typing import Iterable, Mapping, Union, Type, Optional, Any, Sequence, Callable
import itertools
import collections.abc
import functools
import enum
import numbers
import re
import base64
import dataclasses

import numpy as np

from ...jamming import BaseJamDatabase, Jam, JamId, JamItem, JamKind, JamParchment
from marley.utils import ImmutableDict
from ..core import Swank, SwankDatabase
from .base_field import BaseField
from .. import utils



@functools.cache
def _dtype_to_savvy_content(dtype: np.dtype) -> tuple:
    savvy_content = (eval(re.fullmatch(f'dtype\((.+)\)', repr(dtype)).group(1)),) # God forgive me.
    assert np.dtype(*savvy_content) == dtype
    return savvy_content


def object_to_savvy_content(o: Any) -> tuple:
    if hasattr(o, 'to_savvy_content'):
        return o.to_savvy_content()
    elif isinstance(o, (tuple, set, frozenset)):
        return (list(o),)
    elif isinstance(o, (dict, ImmutableDict)):
        return (list(map(list, o.items())),)
    elif dataclasses.is_dataclass(type(o)):
        return tuple(getattr(o, field_name) for field_name in o.__dataclass_fields__)
    elif isinstance(o, np.dtype):
        return _dtype_to_savvy_content(o)
    elif isinstance(o, bytes):
        return (base64.b64encode(o).decode(),)
    elif isinstance(o, type):
        return (utils.type_to_name(o),)
    else:
        assert isinstance(o, enum.Enum)
        return (o.value,)


class SavvyField(BaseField):

    field_type_name = 'savvy'

    def __init__(self, default_factory: Optional[Callable[[], Any]] = type(None)) -> None:
        self.default_factory = default_factory


    def from_jam(self, jam: Jam, swank_database: SwankDatabase, *,
                 swank_cache: dict) -> Optional[Swank]:
        if isinstance(jam, (numbers.Number, str, type(None), bool)):
            return jam
        elif isinstance(jam, list):
            return [self.from_jam(item, swank_database, swank_cache=swank_cache) for item in jam]
        else:
            assert isinstance(jam, dict)
            ((name, savvy_content),) = jam.items()
            assert isinstance(savvy_content, list)
            type_ = utils.name_to_type(name)
            if issubclass(type_, Swank):
                jam_id_name, jam_index = savvy_content
                pair = (JamId(jam_id_name), jam_index)
                try:
                    return swank_cache[pair]
                except KeyError:
                    swank_cache[pair] = swank_cache = swank_database.load_swank(type_, *pair)
                    return swank_cache
            elif type_ is bytes:
                return base64.b64decode(*savvy_content)
            elif issubclass(type_, type):
                result = utils.name_to_type(savvy_content[0])
                assert type(result) is type_
                return result
            else:
                return type_(*(self.from_jam(item, swank_database,
                                             swank_cache=swank_cache) for item in savvy_content))

    def to_jam(self, value: Any, swank_database: SwankDatabase) -> Jam:
        if isinstance(value, (numbers.Number, str, type(None), bool)):
            return value
        elif isinstance(value, list):
            return [self.to_jam(item, swank_database) for item in value]
        else:
            if isinstance(value, Swank):
                if not value.has_jam_id_and_index:
                    value.save()
                assert value.has_jam_id_and_index
            return {
                utils.type_to_name(type(value)):
                    [self.to_jam(item, swank_database) for item in object_to_savvy_content(value)]
            }




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from .int_crowding import IntCrowd, Interval, Point
from .gaining import ThinGain, FatGain
from .jobbing import Job, ThinJob, FatJob, SerialJob, ParallelJob
from .supervising import Supervisor
from .sharking import Shark
from .weighting import Weight, CalfWeight, TadpoleWeight


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import functools
import warnings
import re
import json
import urllib.parse
import threading
import pathlib
import requests
import logging
from typing import (Optional, Tuple, Union, Container, Hashable, Iterator,
                    Iterable, Any, Dict, FrozenSet, Callable, Type, Sequence)

import flask

from marley import constants
from marley.worlds.grid_royale.core import get_games_folder
from marley.jamswank.jamming import server as jamming_server
from marley.jamswank.jamming.jam_file_database import JamFileDatabase
from marley.jamswank import SingletonSwank, SimpleField


class ArezzoUserSession(SingletonSwank):
    favorites = SimpleField(lambda: [])



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

'''A collection of general-purpose tools.'''

from __future__ import annotations

import tempfile
import shutil
import io
import sys
import numbers
import pathlib
import contextlib
import builtins
import collections
import time as time_module
import threading
import copyreg
import contextlib
from typing import Iterable, Iterator, Hashable, Any, Union, Optional
import time
import datetime as datetime_module

from bidict import bidict
from immutabledict import immutabledict as ImmutableDict




@contextlib.contextmanager
def create_temp_folder(prefix: str = tempfile.template, suffix: str = '',
                       parent_folder: Optional[str] = None,
                       chmod: Optional[str] = None) -> pathlib.Path:
    '''
    Context manager that creates a temporary folder and deletes it after usage.

    After the suite finishes, the temporary folder and all its files and
    subfolders will be deleted.

    Example:

        with create_temp_folder() as temp_folder:

            # We have a temporary folder!
            assert temp_folder.is_dir()

            # We can create files in it:
            (temp_folder / 'my_file').open('w')

        # The suite is finished, now it's all cleaned:
        assert not temp_folder.exists()

    Use the `prefix` and `suffix` string arguments to dictate a prefix and/or a
    suffix to the temporary folder's name in the filesystem.

    If you'd like to set the permissions of the temporary folder, pass them to
    the optional `chmod` argument, like this:

        create_temp_folder(chmod=0o550)

    '''
    temp_folder = pathlib.Path(tempfile.mkdtemp(prefix=prefix, suffix=suffix,
                                                dir=parent_folder))
    try:
        if chmod is not None:
            temp_folder.chmod(chmod)
        yield temp_folder
    finally:
        shutil.rmtree(str(temp_folder))


class SelfLoggingDict(collections.UserDict):

    def __setitem__(self, key: Any, value: Any) -> None:
        collections.UserDict.__setitem__(self, key, value)
        print(f'd[{key}] = {value}')

    def __delitem__(self, key: Any) -> None:
        collections.UserDict.__delitem__(self, key)
        print(f'del d[{key}]')


class SelfLoggingBidict(bidict):
    _fwdm_cls = SelfLoggingDict



class LruDict(collections.OrderedDict):

    def __init__(self, max_size: int, /, *args, **kwargs):
        assert max_size > 0
        self.max_size = max_size

        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.move_to_end(key)

        while len(self) > self.max_size:
            del self[next(iter(self))]

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        finally:
            self.move_to_end(key)

    def peek_last_item(self):
        try:
            last_key = next(reversed(self))
        except StopIteration as stop_iteration:
            raise IndexError from stop_iteration
        return (last_key, self[last_key])



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import datetime as datetime_module
from typing import Iterable, Mapping, Union, Type, Optional
import itertools

from ..jamming import BaseJamDatabase, Jam, JamId, JamItem, JamKind, JamParchment
from .core import *
from .fields import *


# Usage: (Will be moved out)

class ProgramRun(Swank):
    creation_datetime: datetime_module.datetime


class Experiment(Swank):
    creation_datetime: datetime_module.datetime

class Game(Swank):
    ...


class BlackjackGame(Game):
    policies: Mapping[str, Policy]


class GridRoyaleSwankDatabase(SwankDatabase):
    ...


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar)
import itertools
import dataclasses
import logging

import more_itertools
import operator as operator_module

from . import int_crowding
from .int_crowding import IntCrowd, Interval, Point

logger = logging.getLogger(__name__)


GainArg = TypeVar('GainArg', bound=Union[IntCrowd, bool, None])

class CantWorkOnEmptyGain(Exception):
    pass



class Gain(abc.ABC):

    def __init__(self, job: Job, arg: GainArg) -> None:
        self.job = job
        self.arg = arg

    @abc.abstractmethod
    def _reduce(self) -> tuple:
        raise NotImplementedError

    @abc.abstractmethod
    def get_doable(self) -> Gain:
        raise NotImplementedError

    @abc.abstractmethod
    def __bool__(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self) -> Optional[Gain]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_gain_desired_by_self(self) -> Gain:
        '''


        We're assuming that `self` is the desired gain, and we want the gain we desire because of
        that. In other words, this exists for `SerialGain` to demand everything from zero to the
        maximum.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_gain_unblocked_by_finished_gain_of_same_job(self, finished_gain: Gain) -> Gain:
        '''


        We're assuming that `self` is the desired gain, and we're looking for unblocked gain that
        intersects with this desired gain.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def get_gain_unblocked_by_finished_parent_gain(self, finished_parent_gain: Gain,
                                                  parent_weight: Weight) -> Gain:
        '''


        We're assuming that `self` is the desired gain, and we're looking for unblocked gain that
        intersects with this desired gain.
        '''
        raise NotImplementedError

    def __eq__(self, other: Any) -> bool:
        return (type(other) == type(self)) and (self._reduce() == other._reduce())

    def __hash__(self) -> int:
        return hash(self._reduce())

    def __repr__(self) -> str:
        return f'{type(self).__name__}{repr(self._reduce()[1:])}'

    @abc.abstractmethod
    def __contains__(self, other: Gain) -> bool:
        raise NotImplementedError


    @abc.abstractmethod
    def __and__(self, other: Union[Gain, Any]) -> Gain:
        raise NotImplementedError

    @abc.abstractmethod
    def __or__(self, other: Union[Gain, Any]) -> Gain:
        raise NotImplementedError

    @abc.abstractmethod
    def __and__(self, other: Union[Gain, Any]) -> Gain:
        raise NotImplementedError

    @abc.abstractmethod
    def __sub__(self, other: Union[Gain, Any]) -> Gain:
        raise NotImplementedError




class ThinGain(Gain):
    def __init__(self, job: ThinJob, is_finished: Optional[bool] = False) -> None:
        self.is_finished = bool(is_finished)
        Gain.__init__(self, job, self.is_finished)

    def _reduce(self) -> tuple:
        return (type(self), self.job, self.is_finished)

    def get_doable(self) -> ThinGain:
        return self

    def __bool__(self) -> bool:
        return self.is_finished

    def get_gain_desired_by_self(self) -> ThinGain:
        return self

    def get_gain_unblocked_by_finished_gain_of_same_job(self, finished_gain: ThinGain) -> ThinGain:
        return self

    def get_gain_unblocked_by_finished_parent_gain(self, finished_parent_gain: Gain,
                                                  parent_weight: Weight) -> Gain:
        if isinstance(finished_parent_gain, ThinGain):
            assert isinstance(parent_weight, CalfWeight)
            assert parent_weight.offset == 0
            return self
        else:
            assert isinstance(finished_parent_gain, FatGain)
            assert isinstance(parent_weight, TadpoleWeight)
            return ThinGain(
                self.job,
                Interval[0 : parent_weight.amount] in finished_parent_gain.int_crowd
            )

    def __call__(self) -> Optional[ThinGain]:
        logger.debug(f'Executing {self}')
        if self:
            return_value = self.job.thin_run()
            logger.debug(f'Finished executing {self}')
            return return_value


    def sniff(self) -> ThinGain:
        if not self:
            return self
        result = self.job.thin_sniff()
        if result is None or isinstance(result, bool):
            return self.job.create_gain(bool(result))
        else:
            assert isinstance(result, ThinGain)
            assert result.job == self.job
            return result

    def __contains__(self, other: ThinGain) -> bool:
        assert isinstance(other, ThinGain) and (self.job == other.job)
        return bool(self or not other)


    def __or__(self, other: Union[Gain, Any]) -> Gain:
        if (type(self) is type(other)) and (self.job == other.job):
            return type(self)(self.job, self.is_finished or other.is_finished)
        else:
            return NotImplemented

    def __and__(self, other: Union[Gain, Any]) -> Gain:
        if (type(self) is type(other)) and (self.job == other.job):
            return type(self)(self.job, self.is_finished and other.is_finished)
        else:
            return NotImplemented

    def __sub__(self, other: Union[Gain, Any]) -> Gain:
        if (type(self) is type(other)) and (self.job == other.job):
            return type(self)(self.job, self.is_finished and not other.is_finished)
        else:
            return NotImplemented


class FatGain(Gain):
    def __init__(self, job: FatJob, int_crowd: Optional[IntCrowd] = None) -> None:
        self.int_crowd: IntCrowd = int_crowd if (int_crowd is not None) else IntCrowd()
        Gain.__init__(self, job, self.int_crowd)


    def _reduce(self) -> tuple:
        return (type(self), self.job, self.int_crowd)

    def get_doable(self) -> FatGain:
        if not self:
            raise CantWorkOnEmptyGain
        first_interval = more_itertools.first(self.int_crowd.intervals)
        first_point = more_itertools.first(first_interval)
        interesting_interval = Interval(first_point[:-1] + first_interval.as_tuple()[-1:])
        money_item = interesting_interval.as_tuple()[-1]
        if isinstance(money_item, int):
            result_interval = interesting_interval
        else:
            start = money_item.start
            end = money_item.stop
            best_end = min(end, start + self.job.max_chunk_size)
            result_interval = Interval(interesting_interval.as_tuple()[:-1] + ((start, best_end),))

        return self.job.create_gain(IntCrowd(result_interval))

    def get_gain_unblocked_by_finished_parent_gain(self, finished_parent_gain: FatGain,
                                                   parent_weight: Weight) -> Gain:
        return parent_weight.get_unblocked_child_gain(self, finished_parent_gain)


    def __call__(self) -> Optional[FatGain]:
        logger.debug(f'Executing {self}')
        if self:
            return_value = self.job.fat_run(self)
            logger.debug(f'Finished executing {self}')
            return return_value

    def sniff(self) -> FatGain:
        result = self.job.fat_sniff(self)
        if result is None:
            return self.job.create_gain()
        elif isinstance(result, IntCrowd):
            return self.job.create_gain(result)
        else:
            assert isinstance(result, FatGain)
            assert result.job == self.job
            return result




    def __contains__(self, other: FatGain) -> bool:
        assert isinstance(other, FatGain) and (self.job == other.job)
        return other.int_crowd in self.int_crowd

    def __bool__(self) -> bool:
        return bool(self.int_crowd)


    def _arithmetic(self, other: Union[IntCrowd, Any], operator: Callable) -> Gain:
        if (type(self) is type(other)) and (self.job == other.job):
            return type(self)(self.job, operator(self.arg, other.arg))
        else:
            return NotImplemented

    def __or__(self, other: Union[Gain, Any]) -> Gain:
        return self._arithmetic(other, operator_module.or_)

    def __and__(self, other: Union[Gain, Any]) -> Gain:
        return self._arithmetic(other, operator_module.and_)

    def __sub__(self, other: Union[Gain, Any]) -> Gain:
        return self._arithmetic(other, operator_module.sub)


class SerialGain(FatGain):
    def get_gain_desired_by_self(self) -> ThinGain:
        if not self:
            return self
        return self.job.create_gain(
            IntCrowd(
                Interval(interval.as_tuple()[:-1] + ((0, interval.sliceful_tuple[-1].stop),))
                for interval in self.int_crowd.intervals
            )
        )

    def get_gain_unblocked_by_finished_gain_of_same_job(self, finished_gain: Gain) -> Gain:
        return SerialGain(
            self.job,
            IntCrowd(
                (point for point in self.int_crowd if (point[-1] == 0) or
                 Point(point[:-1] + ((point[-1] - 1),)) in finished_gain.int_crowd)
            )
        )



class ParallelGain(FatGain):
    def get_gain_desired_by_self(self) -> ParallelGain:
        return self

    def get_gain_unblocked_by_finished_gain_of_same_job(self, finished_gain: Gain) -> Gain:
        return self


@dataclasses.dataclass(order=True, frozen=True)
class GainDyad:
    requested_gain: Gain
    returned_gain: Optional[Gain]




from .weighting import Weight, CalfWeight, TadpoleWeight
from .jobbing import Job, ThinJob, FatJob



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import random
import weakref
import concurrent.futures
import numbers
import functools
import collections.abc
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Mapping, MutableMapping)
import hashlib
import weakref
import dataclasses
import pathlib
import tempfile

import more_itertools
import numpy as np

from marley import gamey
from marley.jamswank.swanking import SwankDatabase, Swank
from . import utils


class GameySwank(Swank):
    def __init__(self, *args, swank_database: Optional[SwankDatabase] = None, **kwargs) -> None:
        if swank_database is not None:
            assert swank_database is gamey.global_gamey_swank_database
        Swank.__init__(self, *args, swank_database=gamey.global_gamey_swank_database, **kwargs)


class GameySwankDatabase(SwankDatabase):
    pass








# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import datetime as datetime_module
from typing import Iterable, Mapping, Union, Type, Optional, Any
import itertools
import functools
import filelock
import re
import collections.abc
import contextlib


from ..jamming import BaseJamDatabase, Jam, JamId, JamItem, JamKind, JamParchment, JamFileDatabase
from . import utils
from .exceptions import EmptyJam
from ..file_locking import FileLock
from .core import Swank, SwankType, SwankDatabase


class SingletonSwankType(SwankType):
    @property
    def _singleton_jam_id(cls) -> JamId:
        return JamId.create(block_size=cls.default_block_size, zero=True)



class SingletonSwank(Swank, metaclass=SingletonSwankType):
    def __init__(self, *, swank_database: SwankDatabase, **kwargs):
        self._singleton_jam_id = type(self)._singleton_jam_id

        # Being forgiving yet watchful if wer're given redundant arguments:
        assert kwargs.pop('jam_id', self._singleton_jam_id) == self._singleton_jam_id
        assert kwargs.pop('jam_index', 0) == 0

        Swank.__init__(
            self,
            swank_database=swank_database,
            jam_id=self._singleton_jam_id,
            jam_index=0,
            **kwargs,
        )

    @classmethod
    def load(cls, swank_database: SwankDatabase) -> SingletonSwank:
        return swank_database.load_swank(cls, jam_id=cls._singleton_jam_id, jam_index=0)

    @classmethod
    def load_or_create(cls, swank_database: SwankDatabase) -> SingletonSwank:
        try:
            return cls.load(swank_database)
        except EmptyJam:
            return cls(swank_database=swank_database)

    @classmethod
    @contextlib.contextmanager
    def lock_and_load_or_create(cls, swank_database: SwankDatabase, *,
                                save: bool = False) -> SingletonSwank:
        jam_kind = swank_database.get_jam_kind(cls)
        lock_path = jam_kind[cls._singleton_jam_id]._get_lock_path()
        parchment_lock = FileLock(lock_path)

        with parchment_lock:
            try:
                singleton_swank = cls.load(swank_database)
            except IndexError:
                singleton_swank = cls(swank_database=swank_database)
            yield singleton_swank
            if save:
                singleton_swank.save()

    def reload(self) -> SingletonSwank:
        return type(self).load(self.swank_database)




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations
from __future__ import generator_stop

from typing import Optional, Union, Tuple, Iterable, Iterator, Sequence

import more_itertools


def iterate_windowed_pairs(iterable: Iterable) -> Iterator:
    iterator = iter(iterable)
    try:
        old = next(iterator)
    except StopIteration:
        return
    for new in iterator:
        yield (old, new)
        old = new


class LastDetectingIterator(more_itertools.peekable):
    def on_last_item(self) -> bool:
        try:
            self.peek()
        except StopIteration:
            return True
        else:
            return False





# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from typing import Iterable, Sequence, Iterable
import itertools
import abc
import datetime as datetime_module
from typing import Iterable, Mapping, Union, Type, Optional, Any
import itertools
import collections.abc
import re
import functools
import types
import numpy as np


def iterable_length(iterable: Iterable) -> int:
    for i, _ in enumerate(iterable):
        pass
    return i + 1


def removesuffix(self: str, suffix: str, /) -> str:
    # Backport of `str.removesuffix` from Python 3.9.
    if suffix and self.endswith(suffix):
        return self[:-len(suffix)]
    else:
        return self[:]


def sliced(seq: Sequence, n: int) -> Iterable[Sequence]:
    """Yield slices of length *n* from the sequence *seq*.

        >>> list(sliced((1, 2, 3, 4, 5, 6), 3))
        [(1, 2, 3), (4, 5, 6)]

    If the length of the sequence is not divisible by the requested slice
    length, the last slice will be shorter.

        >>> list(sliced((1, 2, 3, 4, 5, 6, 7, 8), 3))
        [(1, 2, 3), (4, 5, 6), (7, 8)]

    This function will only work for iterables that support slicing.
    For non-sliceable iterables, see :func:`chunked`.

    """
    # Taken from more_itertools
    return itertools.takewhile(len, (seq[i : i + n] for i in itertools.count(0, n)))


_address_pattern = re.compile(
    r"^(?P<address>([a-zA-Z_][0-9a-zA-Z_]*)(\.[a-zA-Z_][0-9a-zA-Z_]*)*)$"
)
'''Pattern for Python addresses, like 'email.encoders'.'''



def _get_parent_and_dict_from_namespace(namespace):
    '''
    Extract the parent object and `dict` from `namespace`.

    For the `namespace`, the user can give either a parent object
    (`getattr(namespace, address) is obj`) or a `dict`-like namespace
    (`namespace[address] is obj`).

    Returns `(parent_object, namespace_dict)`.
    '''

    if hasattr(namespace, '__getitem__') and hasattr(namespace, 'keys'):
        parent_object = None
        namespace_dict = namespace

    else:
        parent_object = namespace
        namespace_dict = vars(parent_object)

    return (parent_object, namespace_dict)


def is_address(string):
    return bool(_address_pattern.match(string))


def import_if_exists(module_name):
    '''
    Import module by name and return it, only if it exists.
    '''
    try:
        return __import__(module_name)
    except ModuleNotFoundError:
        return None

def cute_import(module_name):
    '''
    Import a module.

    This function has several advantages over `__import__`:

     1. It avoids the weird `fromlist=['']` that you need to give `__import__`
        in order for it to return the specific module you requested instead of
        the outermost package, and

     2. It avoids a weird bug in Linux, where importing using `__import__` can
        lead to a `module.__name__` containing two consecutive dots.

    '''
    if '.' in module_name:
        package_name, submodule_name = module_name.rsplit('.', 1)
        package = __import__(module_name)
        return functools.reduce(getattr,
                                [package] + module_name.split('.')[1:])
    else:
        return __import__(module_name)




def get_object_by_address(address, root=None, namespace={}):
    r'''
    Get an object by its address.

    For example:

        >>> get_object_by_address('email.encoders')
        <module 'email.encoders' from 'c:\Python27\lib\email\encoders.pyc'>

    `root` is an object (usually a module) whose attributes will be looked at
    when searching for the object. `namespace` is a `dict` whose keys will be
    searched as well.
    '''
    # todo: should know what exception this will raise if the address is bad /
    # object doesn't exist.

    assert is_address(address)

    if not _address_pattern.match(address):
        raise ValueError("'%s' is not a legal address." % address)

    ###########################################################################
    # Before we start, we do some pre-processing of `root` and `namespace`:   #

    # We are letting the user input (base)strings for `root` and `namespace`,
    # so if he did that, we'll get the actual objects.

    if root:
        # First for `root`:
        if isinstance(root, str):
            root = get_object_by_address(root)
        root_short_name = root.__name__.rsplit('.', 1)[-1]

    if namespace not in (None, {}):
        # And then for `namespace`:
        if isinstance(namespace, str):
            namespace = get_object_by_address(namespace)

        parent_object, namespace_dict = _get_parent_and_dict_from_namespace(namespace)
    else:
        parent_object, namespace_dict = None, None


    # Finished pre-processing `root` and `namespace`.                         #
    ###########################################################################


    ###########################################################################
    # The implementation is recursive: We handle the case of a single-level
    # address, like 'email'. If we get a multi-level address (i.e. contains a
    # dot,) like 'email.encoders', we use this function twice, first to get
    # `email`, and then from it to get `email.encoders`.

    if '.' not in address:

        ### Here we solve the basic case of a single-level address: ###########
        #                                                                     #

        # Let's rule out the easy option that the requested object is the root:
        if root and (address == root_short_name):
            return root

        if parent_object is not None:

            if isinstance(parent_object, types.ModuleType) and hasattr(parent_object, '__path__'):

                # `parent_object` is a package. The wanted object may be a
                # module. Let's try importing it:

                import_if_exists('.'.join((parent_object.__name__, address)),)
                # Not keeping reference, just importing so we could get later.

        # We know we have a `namespace_dict` to take the object from, and we
        # might have a `parent_object` we can take the object from by using
        # `getattr`. We always have a `namespace_dict`, but not always a
        # `parent_object`.
        #


        # We are going to prefer to do `getattr` from `parent_object`, if one
        # exists, rather than using `namespace_dict`. This is because some
        # attributes may not be present on an object's `__dict__`, and we want
        # to be able to catch them:

        # The first place we'll try to take the object from is the
        # `parent_object`. We try this before `namespace_dict` because
        # `parent_object` may have `__getattr__` or similar magic and our
        # object might be found through that:
        if (parent_object is not None) and hasattr(parent_object, address):
            return getattr(parent_object, address)

        # Next is the `namespace_dict`:
        elif namespace_dict and (address in namespace_dict):
            return namespace_dict[address]

        # Last two options:
        else:
            try:
                # It may be a built-in:
                return eval(address)
            except Exception:
                # Or a module:
                return cute_import(address)

        #                                                                     #
        ### Finished solving the basic case of a single-level address. ########


    else: # '.' in address

        ### If we get a composite address, we solve recursively: ##############
        #                                                                     #

        first_object_address, second_object_address = address.rsplit('.', 1)

        first_object = get_object_by_address(first_object_address, root=root, namespace=namespace)

        second_object = get_object_by_address(second_object_address, namespace=first_object)

        return second_object

        #                                                                     #
        ### Finished solving recursively for a composite address. #############


@functools.cache
def type_to_name(type_: Type, *, safety_check: bool = True) -> str:
    assert isinstance(type_, type)

    # Special handling for Numpy's overly clever _DTypeMeta metaclass:
    if type_.__name__.startswith('dtype') and (type(type_) is type(np.dtype)):
        return 'numpy.dtype'

    name = (type_.__name__ if type_.__module__ == 'builtins'
            else f'{type_.__module__}.{type_.__name__}')

    if safety_check:
        assert name_to_type(name, safety_check=False) is type_
    return name


@functools.cache
def name_to_type(name: str, *, safety_check: bool = True) -> Type:
    type_ = get_object_by_address(name)
    if safety_check:
        assert type_to_name(type_, safety_check=False) == name
    return type_




# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict)
import threading
import functools
import collections
import io
import itertools
import random
import operator as operator_module
import uuid as uuid_module

import more_itertools

from .weighting import Weight, CalfWeight, TadpoleWeight
from .gaining import Gain, ThinGain, FatGain, GainArg, GainDyad
from .jobbing import Job, ThinJob, FatJob, JobSniffingJob
from .wedging import Wedge
from . import utils

import networkx as nx

fields = (
    'job_to_finished_gain',
    'job_to_pending_gain',
    'job_to_preexisting_gain',
    'job_to_sniffed_gain',
    'job_to_sniff_pending_gain',
    'job_to_directive_gain',
    'job_to_desired_gain',
    'job_to_available_gain',
)

def subtract_dicts(minuend_dict: dict, subtrahend_dict: dict) -> dict:
    return {job: gain for job in {*minuend_dict.keys(), *subtrahend_dict.keys()}
            if (gain := (minuend_dict[job] - subtrahend_dict[job]))}



class AntillesDiff:
    def __init__(self, old_antilles_snapshot: AntillesSnapshot,
                 new_antilles_snapshot: AntillesSnapshot) -> None:
        assert old_antilles_snapshot.keys() == new_antilles_snapshot.keys()

        self.old_antilles_snapshot = old_antilles_snapshot
        self.new_antilles_snapshot = new_antilles_snapshot

        self.additions = {
            field: d for field in old_antilles_snapshot
             if (d := subtract_dicts(new_antilles_snapshot[field], old_antilles_snapshot[field]))
        }
        self.subtractions = {
            field: d for field in old_antilles_snapshot
             if (d := subtract_dicts(old_antilles_snapshot[field], new_antilles_snapshot[field]))
        }


    def __repr__(self) -> str:
        return (f'<{type(self).__name__}: '
                f'{len(self.additions)} addition{"s" if len(self.additions) else ""} and '
                f'{len(self.subtractions)} subtraction{"s" if len(self.additions) else ""}.>')

    def __bool__(self) -> bool:
        return bool(self.additions or self.subtractions)

    def as_dict(self) -> dict:
        return {'additions': self.additions, 'subtractions': self.subtractions}

    def show(self) -> str:
        string_io = io.StringIO()
        for d, title in ((self.additions, 'Additions:\n'),
                         (self.subtractions, 'Subtractions:\n')):
            string_io.write(title)
            if d:
                for field, change in d.items():
                    string_io.write(f'    {field}:\n')
                    for job, gain in change.items():
                        string_io.write(f'        {job}: {gain.arg}\n')
            else:
                string_io.write(f'    None.\n')

        return string_io.getvalue()





class AntillesDiffRecorder:
    def __init__(self, antilles: Antilles):
        self.antilles = antilles
        self.is_entered = False
        self.old_antilles_snapshot: Optional[AntillesSnapshot] = None
        self.new_antilles_snapshot: Optional[AntillesSnapshot] = None
        self.antilles_diff: Optional[AntillesDiff] = None


    def __enter__(self) -> AntillesDiffRecorder:
        assert self.is_entered is False
        assert self.old_antilles_snapshot is None
        assert self.new_antilles_snapshot is None
        assert self.antilles_diff is None

        self.is_entered = True
        self.old_antilles_snapshot = AntillesSnapshot.create_from_antilles(self.antilles)
        return self


    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        assert self.is_entered is True
        assert self.old_antilles_snapshot is not None
        assert self.new_antilles_snapshot is None
        assert self.antilles_diff is None

        self.is_entered = False
        self.new_antilles_snapshot = AntillesSnapshot.create_from_antilles(self.antilles)
        self.antilles_diff = AntillesDiff(self.old_antilles_snapshot,
                                          self.new_antilles_snapshot)





class AntillesSnapshot(utils.CuteUserDict):
    @classmethod
    def create_from_antilles(cls, antilles: Antilles) -> AntillesSnapshot:
        with antilles.lock:
            return AntillesSnapshot(
                ((field, JobToItsGainDict(getattr(antilles, field))) for field in fields)
            )

    def show(self) -> str:
        string_io = io.StringIO()
        for field, d in self.items():
            string_io.write(f'{field}:\n')
            if d:
                for job, gain in d.items():
                    string_io.write(f'    {job}: {gain.arg}\n')
            else:
                string_io.write(f'    None.\n')

        return string_io.getvalue()







from .antillesing import Antilles, JobToItsGainDict


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from .misc import union, intersection, ChangeTracker
from .pre_spawning import PreSpawnProcessPoolExecutor
from .sleeping import sleep_until
from .cute_user_dicting import CuteUserDict


# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import time
import threading
import os
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict, Set)
from types import TracebackType
import concurrent.futures
from bidict import bidict
import logging

import more_itertools

import marley
from . import sharknado
from . import jamswank

def marley_worker_setup_function(swank_database: jamswank.SwankDatabase,
                                 logging_kwargs: dict) -> None:
    from . import gamey
    gamey.global_gamey_swank_database = swank_database

    # Avoid TensorFlow spam:
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
    os.environ['CUDA_VISIBLE_DEVICES'] = '0'
    logging.getLogger('tensorflow').addFilter(
        lambda record: 'Tracing is expensive and the excessive' not in record.msg
    )

    marley.logging_setup.setup(**logging_kwargs)




class MarleyShark(sharknado.Shark):
    def __init__(self,
                 directive_thin_jobs: Union[sharknado.ThinJob, Iterable[sharknado.ThinJob]] = (), *,
                 name: Optional[str] = None, use_multiprocessing: bool = True, start: bool = False,
                 sniff_only: bool = False) -> None:
        sharknado.Shark.__init__(
            self, directive_thin_jobs=directive_thin_jobs, name=name,
            use_multiprocessing=use_multiprocessing, start=start, sniff_only=sniff_only,
            worker_setup_function=marley_worker_setup_function,
            worker_setup_args=(marley.constants.swank_database,
                               marley.logging_setup.get_logging_kwargs())
        )



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import datetime as datetime_module
from typing import Iterable, Mapping, Union, Type, Optional, Any
import itertools
import collections.abc
import re
import functools
import types
import numpy as np

from ..jamming.utils import type_to_name, name_to_type



def camel_case_to_lower_case(s: str) -> str:
    '''
    Convert a string from camel-case to lower-case.

    Example:

        camel_case_to_lower_case('HelloWorld') == 'hello_world'

    '''
    return re.sub(r'(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))', r'_\1', s).lower().strip('_')

